import unittest

import rubin_scheduler.scheduler.detailers as detailers
from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.utils import ObservationArray


class TestDetailers(unittest.TestCase):

    def test_random_filter(self):
        obs = ObservationArray(1)
        obs["filter"] = "r"

        det = detailers.RandomFilterDetailer(filters="iz")

        conditions = Conditions()
        conditions.night = 3
        conditions.mounted_filters = ["i", "z"]

        out_obs = det(obs, conditions)
        assert (out_obs["filter"] == "i") | (out_obs["filter"] == "z")

        # Check that we fall back properly
        conditions.mounted_filters = ["r", "g", "u", "y"]
        det = detailers.RandomFilterDetailer(filters="iz", fallback_order="y")
        out_obs = det(obs, conditions)

        assert out_obs["filter"] == "y"


if __name__ == "__main__":
    unittest.main()



import unittest

import pandas as pd

import rubin_scheduler.scheduler.basis_functions as basis_functions
import rubin_scheduler.scheduler.surveys as surveys
from rubin_scheduler.scheduler.example import simple_greedy_survey
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler
from rubin_scheduler.scheduler.utils import ObservationArray


class TestCoreSched(unittest.TestCase):
    def testsched(self):

        # Just set up a very simple survey, one filter
        survey = simple_greedy_survey(filtername="r")

        scheduler = CoreScheduler([survey])

        observatory = ModelObservatory()

        # Check that we can update conditions
        scheduler.update_conditions(observatory.return_conditions())

        # Check that we can get an observation out
        obs = scheduler.request_observation()
        assert obs is not None

        # Check that we can flush the Queue
        scheduler.flush_queue()
        assert len(scheduler.queue) == 0

        # Check that we can add an observation
        scheduler.add_observation(obs)

        # Check dunder methods
        self.assertIsInstance(repr(scheduler), str)
        self.assertIsInstance(str(scheduler), str)

        # Check access methods
        _ = scheduler.get_basis_functions([0, 0])
        _ = scheduler.get_healpix_maps([0, 0])

        # Check survey access methods
        reward_df = scheduler.make_reward_df(observatory.return_conditions())
        self.assertIsInstance(reward_df, pd.DataFrame)
        reward_df = scheduler.make_reward_df(observatory.return_conditions(), accum=False)
        self.assertIsInstance(reward_df, pd.DataFrame)

        obs = scheduler.request_observation()
        surveys_df = scheduler.surveys_df(0)
        self.assertIsInstance(surveys_df, pd.DataFrame)

        # Test we can record basis function values when requested
        recording_scheduler = CoreScheduler([survey], keep_rewards=True)
        recording_scheduler.update_conditions(observatory.return_conditions())
        obs = recording_scheduler.request_observation()
        self.assertIsInstance(recording_scheduler.queue_reward_df, pd.DataFrame)

    def test_add_obs(self):
        """Test that add_observation works with slice or array"""

        bfs = [basis_functions.ForceDelayBasisFunction(days_delay=3.0, scheduler_note="survey")]
        survey = surveys.GreedySurvey(bfs, [0.0])
        scheduler = CoreScheduler([survey])

        obs = ObservationArray()
        obs["scheduler_note"] = "survey"
        obs["mjd"] = 100

        scheduler.add_observation(obs)

        recorded = scheduler.survey_lists[0][0].basis_functions[0].survey_features["last_obs_self"].feature

        assert recorded["mjd"] == 100
        assert recorded["scheduler_note"] == "survey"

        bfs = [basis_functions.ForceDelayBasisFunction(days_delay=3.0, scheduler_note="survey")]
        survey = surveys.GreedySurvey(bfs, [0.0])
        scheduler = CoreScheduler([survey])

        # Again, but now add just the row rather than full array
        scheduler.add_observation(obs[0])
        recorded = scheduler.survey_lists[0][0].basis_functions[0].survey_features["last_obs_self"].feature

        assert recorded["mjd"] == 100
        assert recorded["scheduler_note"] == "survey"


if __name__ == "__main__":
    unittest.main()



import importlib.util
import urllib.parse
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cache, cached_property
from math import log, sqrt
from warnings import warn

import healpy as hp
import numpy as np
import pandas as pd

# import rubin_sim
from astropy.coordinates import angular_separation
from astropy.coordinates.earth import EarthLocation
from astropy.time import Time

from rubin_scheduler.scheduler.utils import SchemaConverter
from rubin_scheduler.site_models import Almanac, SeeingModel
from rubin_scheduler.utils import (
    Site,
    SysEngVals,
    _angular_separation,
    _approx_altaz2pa,
    pseudo_parallactic_angle,
    rotation_converter,
    survey_start_mjd,
)

GAUSSIAN_FWHM_OVER_SIGMA: float = 2.0 * sqrt(2.0 * log(2.0))


def query_consdb(query: str, url: str = "postgresql://usdf@usdf-summitdb.slac.stanford.edu:5432/exposurelog"):
    """_summary_

    Parameters
    ----------
    query : `str`
        The SQL query to send to the consdb.
    url : `str`, optional
        The database connection string,
        by default "postgresql://usdf@usdf-summitdb.slac.stanford.edu:5432/exposurelog"

    Returns
    -------
    result : `pandas.DataFrame`
        The result of the query.
    """
    url_scheme: str = urllib.parse.urlparse(url).scheme
    match url_scheme:
        case "postgresql":
            can_support_postgresql: bool = bool(
                importlib.util.find_spec("sqlalchemy") and importlib.util.find_spec("psycopg2")
            )
            if not can_support_postgresql:
                raise RuntimeError("Optional dependencies required for postgresql access not installed")

            # import sqlalchemy here rather than at the top to keep mypy happy.
            # Add the type: ignore so IDEs do not complain when running in an
            # environment without it.
            import sqlalchemy  # type: ignore

            connection = sqlalchemy.create_engine(url)
            query_results: pd.DataFrame = pd.read_sql(query, connection)
        case "http" | "https":
            can_support_http: bool = bool(
                importlib.util.find_spec("lsst")
                and importlib.util.find_spec("lsst.summit")
                and importlib.util.find_spec("lsst.summit.utils")
            )
            if not can_support_http:
                raise RuntimeError("Optional dependencies required for ConsDB access not installed")

            # import ConsDbClient here rather than at the top to satisfy mypy.
            # Add the type: ignore so IDEs do not complain when running in an
            # environment without it.
            from lsst.summit.utils import ConsDbClient  # type: ignore

            consdb = ConsDbClient(url)
            query_results: pd.DataFrame = consdb.query(query).to_pandas()
        case _:
            raise ValueError(f"Unrecongined url scheme {url_scheme} in consdb url {url}")

    return query_results


# Different instruments (and different numbers of snaps per visit) do or
# will have differenc schema in the consdb, so make a base class with
# expected common functionality, and use derived classes for
# instrument-specific elements.

# Separate different pieces of metadata (columns in a table of visits) into
# different methods to make it easy to individually override them by
# instrument.


# cached_properties are a way to compute values when and only when needed
# but these won't update if the date is changed. In my expected use, that's
# fine; the user should just instantiate another ConsDBVisits instance if
# they want a different database or date. But, what if someone does?
# To avoid this, make it a frozen dataclass, which prevents members from
# being updated.
@dataclass(frozen=True)
class ConsDBVisits(ABC):
    """Visits returned from a query to the consdb.

    Parameters
    ----------
    day_obs : `str`
        The date for which to query the database.
    url : `str`
        The connection string from the database.
    """

    day_obs: str | int
    url: str = "postgresql://usdf@usdf-summitdb.slac.stanford.edu:5432/exposurelog"

    @property
    @abstractmethod
    def instrument(self) -> str:
        """The instrument.

        Returns
        -------
        instrument : `str`
            The instrument.
        """
        # Subclasses should return a string constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @property
    @abstractmethod
    def num_exposures(self) -> int:
        """The number of snaps per visit.

        Returns
        -------
        num_exposures : `int`
            The number of snaps per visit.
        """
        # Subclasses should return a string constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @property
    @abstractmethod
    def telescope(self) -> str:
        """The telescope.

        Returns
        -------
        telescope : `str`
            The telescope.
        """
        # Subclasses should return a constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @property
    @abstractmethod
    def gain(self) -> pd.Series:
        """The gain.

        Returns
        -------
        gain : `pd.Series`
            The gain.
        """
        # Subclasses should return a constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @property
    @abstractmethod
    def pixel_scale(self) -> pd.Series:
        """The pixel scale.

        Returns
        -------
        pixel_scale : `pd.Series`
            The pixel scale.
        """
        # Subclasses should return a constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @abstractmethod
    def instrumental_zeropoint_for_band(self, band: str) -> float:
        """The zero point.

        Parameters
        ----------
        band : `str`
            The band for which to return the zero point.

        Returns
        -------
        zero_point : `float`
            The photometric zero point.
        """
        # Subclasses should return a constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @property
    @abstractmethod
    def site(self) -> Site:
        """The site.

        Returns
        -------
        site : `rubin_scheduler.utils.site.Site`
            The observatory site.
        """
        # Subclasses should return a constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @cached_property
    def almanac(self) -> Almanac:
        """The astronomical almanac.

        Returns
        -------
        almanac : `rubin_scheduler.site_models.almanac.Almanac`
            Astronomical almanac data.
        """
        return Almanac()

    @cached_property
    def location(self) -> EarthLocation:
        """The observatory location.

        Returns
        -------
        location : `astropy.coordinates.earth.EarthLocation`
            The observatory location.
        """
        #
        return EarthLocation(lat=self.site.latitude, lon=self.site.longitude, height=self.site.height)

    @cached_property
    def day_obs_time(self) -> Time:
        """The day_obs as an astropy time.

        Returns
        -------
        day_obs : `astropy.time.Time`
            The astropy time representation of day_obs.
        """
        time: Time = Time(self.day_obs)
        return time

    @cached_property
    def day_obs_mjd(self) -> int:
        """The day_obs as an mjd.

        Returns
        -------
        day_obs_mjd : `int`
            The mjd representation of day_obs.
        """
        # make mypy happy
        assert isinstance(self.day_obs_time.mjd, float)
        return int(self.day_obs_time.mjd)

    @cached_property
    def day_obs_int(self) -> int:
        """The day_obs as an integer.

        Returns
        -------
        day_obs_int : `int`
            The integer representation of day_obs.
        """
        # make mypy happy
        assert isinstance(self.day_obs_time.iso, str)
        return int(self.day_obs_time.iso[:10].replace("-", ""))

    @cached_property
    def consdb_visits(self) -> pd.DataFrame:
        """The visit data as returned from the consdb

        Returns
        -------
        consdb_visits : `pd.DataFrame`
            The visit data.
        """
        # In the schema as of now, all visits are one exposure,
        # and visit_id = exposure_id.
        # To support 2 exposures snaps in the future, this query will
        # need to be rewritten to group by visit_id.
        consdb_visits_query: str = f"""
            SELECT * FROM cdb_{self.instrument}.exposure AS e
            LEFT JOIN cdb_{self.instrument}.visit{self.num_exposures}_quicklook
                AS v{self.num_exposures}q ON e.exposure_id = v{self.num_exposures}q.visit_id
            WHERE e.obs_start_mjd IS NOT NULL
                AND e.s_ra IS NOT NULL
                AND e.s_dec IS NOT NULL
                AND e.sky_rotation IS NOT NULL
                AND ((e.band IS NOT NULL) OR (e.physical_filter IS NOT NULL))
                AND e.day_obs = {self.day_obs_int}
        """
        return query_consdb(consdb_visits_query, self.url)

    @cached_property
    def visit_id(self) -> pd.Series:
        """The unique identifier from the consdb database.

        Returns
        -------
        visit_id : `pd.Series`
            The unique identifier from the consdb database.
        """
        return self.consdb_visits["visit_id"]

    @cached_property
    def ra(self) -> pd.Series:
        """The R.A. of each visit, in degrees.

        Returns
        -------
        ra : `pd.Series`
            The R.A. of each visit, in degrees.
        """
        return self.consdb_visits["s_ra"]

    @cached_property
    def decl(self) -> pd.Series:
        """The declination if each visit, in degrees.

        Returns
        -------
        decl : `pd.Series`
            The declination of each visit, in degrees.
        """
        return self.consdb_visits["s_dec"]

    @cached_property
    def obs_start_mjd(self) -> pd.Series:
        """The MJD at the start of each visit.

        Returns
        -------
        obs_start_mjd : `pd.Series`
            The MJD at the start of each visit.
        """
        return self.consdb_visits["obs_start_mjd"]

    @cached_property
    def shut_time(self) -> pd.Series:
        """Spatially-averaged shutter-open duration of each visit.

        Returns
        -------
        shut_time : `pd.Series`
            Spatially-averaged shutter-open duration.
        """
        return self.consdb_visits["shut_time"]

    @cached_property
    def obs_midpt_mjd(self) -> pd.Series:
        """Midpoint time for exposure at the fiducial center of the focal
        plane. array in MJD.

        Returns
        -------
        midpt_mjd: `pd.Series`
            Midpoint time for exposure at the fiducial center of the focal
            plane. array in MJD.
        """
        return self.consdb_visits["obs_midpt_mjd"]

    @cached_property
    def sky_rotation(self) -> pd.Series:
        """Targeted sky rotation angle.

        Returns
        -------
        sky_rotation : `pd.Series`
            Targeted sky rotation angle for each visit.
        """
        return self.consdb_visits["sky_rotation"]

    @cached_property
    def target_name(self) -> pd.Series:
        """The target name for each visit.

        Returns
        -------
        target_name : `pd.Series`
            The target name for each visit.
        """
        return self.consdb_visits["target_name"]

    @cached_property
    def airmass(self) -> pd.Series:
        """Airmass of each visit.

        Returns
        -------
        airmass : `pd.Series`
            Airmass of each visit.
        """
        return self.consdb_visits["airmass"]

    @cached_property
    def band(self) -> pd.Series:
        """Abstract filter for each visit

        Returns
        -------
        band : `pd.Series`
            Abstract filter for each visit.
        """
        return self.consdb_visits["band"]

    @cached_property
    def azimuth(self) -> pd.Series:
        """Azimuth for each visit (degrees) at central time.

        Returns
        -------
        azimuth : `pd.Series`
            Azimuth for each visit.
        """
        return self.consdb_visits["azimuth"]

    @cached_property
    def altitude(self) -> pd.Series:
        """Altitude for each visit (degrees) at central time.

        Returns
        -------
        altitude : `pd.Series`
            Altitude for each visit (degrees).
        """
        return self.consdb_visits["altitude"]

    @cached_property
    def zd(self) -> pd.Series:
        """Zenith distance for each visit at central time (degrees)

        Returns
        -------
        zd : `pd.Series`
            Zenith distance for each visit at central time (degrees)
        """
        return 90 - self.altitude

    @cached_property
    def note(self) -> pd.Series:
        """Note for each visit.

        Returns
        -------
        note : `pd.Series`
            Note for each visit.
        """
        return self.consdb_visits["observation_reason"]

    @cached_property
    def visit_time(self) -> pd.Series:
        """Visit time for each visit (seconds).

        Returns
        -------
        visit_time : `pd.Series`
            Visit time for each visit.
        """
        return (self.consdb_visits["obs_end_mjd"] - self.consdb_visits["obs_start_mjd"]) * 24 * 60 * 60

    @cached_property
    def cloud(self) -> pd.Series:
        """Fractional cloud cover of the sky for each visit.

        Returns
        -------
        cloud : `pd.Series`
            Fractional cloud cover of the sky for each visit.
        """
        return pd.Series(np.nan, index=self.consdb_visits.index)

    @cached_property
    def slew_distance(self) -> pd.Series:
        """Slew distance to reach each visit (degrees).

        Returns
        -------
        slew_distance : `pd.Series`
            Slew distance to reach each visit (degrees).
        """
        slew_distance = pd.Series(np.nan, index=self.consdb_visits.index)
        previous_visit_idx = None
        for visit_idx in self.consdb_visits.index:
            if previous_visit_idx is not None:
                this_slew_distance = angular_separation(
                    np.radians(self.ra[previous_visit_idx]),
                    np.radians(self.decl[previous_visit_idx]),
                    np.radians(self.ra[visit_idx]),
                    np.radians(self.decl[visit_idx]),
                )
                slew_distance[visit_idx] = np.degrees(this_slew_distance)
            previous_visit_idx = visit_idx
        return slew_distance

    @cached_property
    def start_time(self) -> Time:
        """Start time for each visit.

        Returns
        -------
        start_time : `Time`
            Start time for each visit.
        """
        return Time(self.obs_start_mjd, format="mjd")

    @cached_property
    def observation_start_lst(self) -> pd.Series:
        """Apparent local sidereal time (degrees) at the start of each visit.

        Returns
        -------
        lst : `pd.Series`
            Apparent local sidereal time (degrees) at the start of each visit.
        """
        start_lst = pd.Series(
            self.start_time.sidereal_time("apparent", self.location).degree, index=self.consdb_visits.index
        )
        return start_lst

    @cached_property
    def pseudo_parallactic_angle(self) -> pd.Series:
        """Parallactic angle as defined in SMTN-019 for each visit (degrees).

        Returns
        -------
        pa : `pd.Series`
            Parallactic angle as defined in SMTN-019 for each visit.
        """
        # Following sim_runner
        # Using pseudo_parallactic_angle, see https://smtn-019.lsst.io/v/DM-44258/index.html

        # make mypy happy
        assert isinstance(self.decl.values, np.ndarray)
        assert isinstance(self.ra.values, np.ndarray)
        assert isinstance(self.obs_start_mjd.values, np.ndarray)
        assert isinstance(self.site.longitude, float)
        assert isinstance(self.site.latitude, float)
        assert isinstance(self.site.height, float)

        pa, alt, az = pseudo_parallactic_angle(
            self.ra.values,
            self.decl.values,
            self.obs_start_mjd.values,
            lon=self.site.longitude,
            lat=self.site.latitude,
            height=self.site.height,
        )
        return pa

    @cached_property
    def rot_tel_pos(self) -> pd.Series:
        """Telescope rotator position (degrees).

        Returns
        -------
        rot_tel_pos : pd.Series
            Telescope rotator position (degrees).
        """
        rot_tel_pos = rotation_converter(telescope=self.telescope)._rotskypos2rottelpos(
            np.radians(self.sky_rotation), np.radians(self.pseudo_parallactic_angle)
        )
        return np.degrees(rot_tel_pos)

    @cached_property
    def parallactic_angle(self) -> pd.Series:
        """Parallactic angle (degrees).

        Returns
        -------
        pa : `pd.Series`
            Parallactic angle (degrees).
        """
        # Following sim_runner
        pa = pd.Series(
            np.degrees(
                _approx_altaz2pa(np.radians(self.altitude), np.radians(self.azimuth), self.site.latitude_rad)
            ),
            index=self.consdb_visits.index,
        )
        return pa

    @cached_property
    def night(self) -> int:
        """The night of the survey.

        Returns
        -------
        night : `int`
            The night of the survey.
        """
        night: int = 1 + self.day_obs_mjd - int(np.floor(survey_start_mjd()))
        return night

    @cached_property
    def sun_moon_positions(self) -> pd.DataFrame:
        """Sun and moon R.A., Declination, Azimuth, and Altitude (degrees)
        at each visit.

        Returns
        -------
        positions : `pd.DataFrame`
            Sun and moon positions (degrees).
        """
        sun_moon = pd.DataFrame(
            self.almanac.get_sun_moon_positions(self.obs_midpt_mjd.values), index=self.consdb_visits.index
        )

        # Convert to degrees, because everything else is in degrees
        for body in ("sun", "moon"):
            for coord in ("alt", "az", "RA", "dec"):
                column = f"{body}_{coord}"
                sun_moon[column] = np.degrees(sun_moon[column])

        return sun_moon

    @cached_property
    def moon_distance(self) -> pd.Series:
        """Angle between each visit pointing the moon (degrees).

        Returns
        -------
        distance : `pd.Series`
            Angle between each visit pointing the moon (degrees).
        """
        # make mypy happy
        assert isinstance(self.decl.values, np.ndarray)
        assert isinstance(self.ra.values, np.ndarray)

        moon_ra = self.sun_moon_positions["moon_RA"].values
        moon_decl = self.sun_moon_positions["moon_dec"].values
        assert isinstance(moon_ra, np.ndarray)
        assert isinstance(moon_decl, np.ndarray)

        moon_distance_rad = _angular_separation(
            np.radians(self.ra.values),
            np.radians(self.decl.values),
            np.radians(moon_ra),
            np.radians(moon_decl),
        )
        return pd.Series(np.degrees(moon_distance_rad), index=self.consdb_visits.index)

    @cached_property
    def solar_elong(self) -> pd.Series:
        """Solar elongation (degrees) of each visit.

        Returns
        -------
        elong : `pd.Series`
            Solar elongation of each visit in degrees.
        """
        # make mypy happy
        assert isinstance(self.decl.values, np.ndarray)
        assert isinstance(self.ra.values, np.ndarray)

        sun_ra = self.sun_moon_positions["sun_RA"].values
        sun_decl = self.sun_moon_positions["sun_dec"].values
        assert isinstance(sun_ra, np.ndarray)
        assert isinstance(sun_decl, np.ndarray)

        solar_elong_rad = _angular_separation(
            np.radians(self.ra.values),
            np.radians(self.decl.values),
            np.radians(sun_ra),
            np.radians(sun_decl),
        )
        return pd.Series(np.degrees(solar_elong_rad), index=self.consdb_visits.index)

    @cached_property
    def zeropoint_e(self) -> pd.Series:
        """Photometric zero point for each visit, with "counts" in electons.

        Returns
        -------
        zp : `pd.Series`
            Photometric zero point for each visit, with "counts" in electons.
        """
        return self.consdb_visits["zero_point_median"] + 2.5 * np.log10(self.gain)

    @cached_property
    def instrumental_zeropoint_e(self) -> pd.Series:
        """Instrumental zero point for each visit, with "counts" in electons.

        Returns
        -------
        zp : `pd.Series`
            Instrumental zero point for each visit, with "counts" in electons.
        """
        return self.band.apply(self.instrumental_zeropoint_for_band) + 2.5 * np.log10(self.shut_time)

    @cached_property
    def sky_e_per_pixel(self) -> pd.Series:
        """Median sky background of each visit (in electons).

        Returns
        -------
        sky : `pd.Series`
            Median sky background of each visit (in electons).
        """
        return self.consdb_visits["sky_bg_median"] * self.gain

    @cached_property
    def sky_mag_per_asec(self) -> pd.Series:
        """Median sky background in each visit, in mags asec^-2

        Returns
        -------
        sky : `pd.Series`
            Median sky background in each visit, in mags asec^-2
        """
        return self.instrumental_zeropoint_e - 2.5 * np.log10(self.sky_e_per_pixel / (self.pixel_scale**2))

    @cache
    def hpix(self, nside: int) -> pd.Series:
        """The heaplix of the pointing for each visit, for a given nside.

        Parameters
        ----------
        nside : `int`
            The nside of the healpix map to use.

        Returns
        -------
        hpix : `pd.Series`
            The heaplix of the pointing for each visit, for a given nside.
        """
        return pd.Series(hp.ang2pix(nside, self.ra, self.decl, lonlat=True), index=self.consdb_visits.index)

    @cached_property
    def fwhm_eff(self) -> pd.Series:
        """Effective PSF FWHM (arcseconds).

        Returns
        -------
        fwhm : `pd.Series`
            Effective PSF FWHM (arcseconds).
        """
        return self.consdb_visits["psf_sigma_median"] * GAUSSIAN_FWHM_OVER_SIGMA * self.pixel_scale

    @cached_property
    def fwhm_geom(self) -> pd.Series:
        """Geometric PSF FWHM (arcseconds).

        Returns
        -------
        fwhm : `pd.Series`
            Geometric PSF FWHM (arcseconds).
        """
        return SeeingModel.fwhm_eff_to_fwhm_geom(self.fwhm_eff)

    @cached_property
    def fwhm_500(self) -> pd.Series:
        """FWHM at zenith and 500nm, in arcseconds, at the time of each visit.

        Returns
        -------
        fwhm : `pd.Series`
            FWHM at zenith and 500nm, in arcseconds, at the time of each visit.
        """
        return self.consdb_visits["seeing_zenith_500nm_median"] * GAUSSIAN_FWHM_OVER_SIGMA * self.pixel_scale

    @cached_property
    def eff_time_m5(self) -> pd.Series:
        """Effective exposure time, in seconds.
        This is the time to reach the m5 depeth in each visit, under
        a reference set of conditions.

        Returns
        -------
        eff_time : `pd.Series`
            The effective exposure time.
        """
        return self.consdb_visits["eff_time_m5"]

    @cached_property
    def opsim(self) -> pd.DataFrame:
        """The table of visits, in a format replicating opsim output.

        Returns
        -------
        opsim : `pd.DataFrame`
            The table of visits, in a format replicating opsim output.
        """

        opsim_df = pd.DataFrame(
            {
                "observationId": self.visit_id,
                "fieldRA": self.ra,
                "fieldDec": self.decl,
                "observationStartMJD": self.obs_start_mjd,
                "flush_by_mjd": np.nan,
                "visitExposureTime": self.shut_time,
                "filter": self.band,
                "rotSkyPos": self.sky_rotation,
                "rotSkyPos_desired": np.nan,
                "numExposures": self.num_exposures,
                "airmass": self.airmass,
                "seeingFwhm500": self.fwhm_500,
                "seeingFwhmEff": self.fwhm_eff,
                "seeingFwhmGeom": self.fwhm_geom,
                "skyBrightness": self.sky_mag_per_asec,
                "night": self.night,
                "slewTime": np.nan,
                "visitTime": self.visit_time,
                "slewDistance": self.slew_distance,
                "fiveSigmaDepth": self.eff_time_m5,
                "altitude": self.altitude,
                "azimuth": self.azimuth,
                "paraAngle": self.parallactic_angle,
                "psudoParaAngle": self.pseudo_parallactic_angle,
                "cloud": self.cloud,
                "moonAlt": self.sun_moon_positions["moon_alt"],
                "sunAlt": self.sun_moon_positions["sun_alt"],
                "note": self.note,
                "target": self.target_name,
                #                "fieldId": None,
                #                "proposalId": None,
                #                "block_id": None,
                "observationStartLST": self.observation_start_lst,
                "rotTelPos": self.rot_tel_pos,
                "rotTelPos_backup": None,
                "moonAz": self.sun_moon_positions["moon_az"],
                "sunAz": self.sun_moon_positions["sun_az"],
                "sunRA": self.sun_moon_positions["sun_RA"],
                "sunDec": self.sun_moon_positions["sun_dec"],
                "moonRA": self.sun_moon_positions["moon_RA"],
                "moonDec": self.sun_moon_positions["moon_dec"],
                "moonDistance": self.moon_distance,
                "solarElong": self.solar_elong,
                "moonPhase": self.sun_moon_positions["moon_phase"],
                #                "cummTelAz": None,
                #                "scripted_id": None,
                "start_date": self.consdb_visits["obs_start"],
                "t_eff": self.consdb_visits["eff_time_median"],
                "seq_num": self.consdb_visits["seq_num"],
            }
        )

        return opsim_df

    @cached_property
    def obs(self) -> np.recarray:
        """The table of visits, in a format replicating an observation array.

        Returns
        -------
        obs : `np.recarray`
            The observation array.
        """
        return SchemaConverter().opsimdf2obs(self.opsim)


class ComcamSimConsDBVisits(ConsDBVisits):

    def _have_numeric_values(self, column) -> bool:
        if column not in self.consdb_visits.columns:
            return False

        # Called to see if we need to guess values because the
        # consdb doesn't yet fill in what we need
        value_dtype = self.consdb_visits[column].dtype

        # Needed to make mypy happy
        assert isinstance(value_dtype, np.dtype)

        have_values = isinstance(value_dtype, np.number)
        if not have_values:
            warn(f"Consdb does not have values for {column}, guessing values instead")

        return isinstance(value_dtype, np.number)

    # Set "constant" properties using cached_property rather
    # than actual attributes, because abstract "members" can only be declared
    # as such using properties, and overriding a property with an actual
    # attribute confuses type checkers.

    @property
    def instrument(self) -> str:
        return "lsstcomcamsim"

    @property
    def num_exposures(self) -> int:
        return 1

    @property
    def telescope(self) -> str:
        return "rubin"

    @property
    def gain(self) -> pd.Series:
        return pd.Series(1.67, index=self.consdb_visits.index)

    @property
    def pixel_scale(self) -> pd.Series:
        return pd.Series(0.2, index=self.consdb_visits.index)

    def instrumental_zeropoint_for_band(self, band: str) -> float:
        return SysEngVals().zp_t[band]

    @property
    def site(self):
        return Site("LSST")

    @cached_property
    def azimuth(self) -> pd.Series:
        if self._have_numeric_values("azimuth"):
            azimuth = super().azimuth
        elif self._have_numeric_values("azimuth_start"):
            if self._have_numeric_values("azimuth_end"):
                azimuth = (self.consdb_visits["azimuth_start"] + self.consdb_visits["azimuth_end"]) / 2
            else:
                azimuth = self.consdb_visits["azimuth_start"]
        else:
            azimuth = pd.Series(np.nan, index=self.consdb_visits.index)

        return azimuth

    @cached_property
    def altitude(self) -> pd.Series:
        if self._have_numeric_values("altitude"):
            altitude = super().altitude
        elif self._have_numeric_values("altitude_start"):
            if self._have_numeric_values("altitude_end"):
                altitude = (self.consdb_visits["altitude_start"] + self.consdb_visits["altitude_end"]) / 2
            else:
                altitude = self.consdb_visits["altitude_start"]
        else:
            altitude = pd.Series(np.nan, index=self.consdb_visits.index)

        return altitude

    @cached_property
    def airmass(self) -> pd.Series:
        if self._have_numeric_values("airmass"):
            airmass = super().airmass
        else:
            # Calculate airmass using Cassini's model.
            # Models the atmosphere with a uniform spherical shell with a
            # height 1/470 of the radius of the earth.
            # Described in Kirstensen https://doi.org/10.1002/asna.2123190313
            # and https://doi.org/10.2172/1574836
            a_cos_zd = 470 * np.cos(np.radians(self.zd))
            airmass = pd.Series(np.sqrt(a_cos_zd**2 + 941) - a_cos_zd, index=self.consdb_visits.index)

        return airmass

    @cached_property
    def shut_time(self) -> pd.Series:
        if self._have_numeric_values("shut_time"):
            shut_time = super().shut_time
        else:
            shut_time = pd.Series(SysEngVals().exptime, index=self.consdb_visits.index)

        return shut_time

    @cached_property
    def obs_midpt_mjd(self) -> pd.Series:
        if self._have_numeric_values("obs_midpt_mjd"):
            mjd = super().obs_midpt_mjd
        elif self._have_numeric_values("obs_end_mjd"):
            mjd = (self.obs_start_mjd + self.consdb_visits["obs_end_mjd"]) / 2
        else:
            mjd = self.obs_start_mjd + (0.5 * self.shut_time) / 86400

        return mjd

    @cached_property
    def eff_time_m5(self):
        if self._have_numeric_values("eff_time_m5"):
            eff_time = super().eff_time_m5
        else:
            # From SMTN-002
            ref_mags = {"u": 23.70, "g": 24.97, "r": 24.52, "i": 24.13, "z": 23.56, "y": 22.55}
            eff_time = self.band.map(ref_mags) + 1.25 * np.log10(self.consdb_visits["eff_time_median"] / 30)

        return eff_time

    @cached_property
    def band(self) -> pd.Series:
        band = self.consdb_visits["band"].copy()
        missing_band: pd.Series[bool] = band.isna()
        band[missing_band] = self.consdb_visits.loc[missing_band, "physical_filter"].str.get(0)
        return band


# Different subclasses of ConsDBOpsimConverter are needed for
# different instruments.
# Make a factory function to return the correct one.
def load_consdb_visits(instrument: str = "lsstcomcamsim", *args, **kwargs) -> ConsDBVisits:
    """Return visits from the consdb.

    Parameters
    ----------
    instrument : `str`, optional
        The name of the instrument, by default "lsstcomcamsim"

    Returns
    -------
    visits : `ConsDBVisits`
        Visits from the consdb.
    """
    match instrument:
        case "lsstcomcamsim":
            converter: ConsDBVisits = ComcamSimConsDBVisits(*args, **kwargs)
        case _:
            raise NotImplementedError

    return converter



import sys

import astropy.units as u
import numpy as np
from astroplan import Observer
from astropy.time import Time

from rubin_scheduler.utils import Site

if __name__ == "__main__":
    # Trying out the astroplan sunrise/set code.
    # conda install -c astropy astroplan
    mjd_start = 59853.5 - 3.0 * 365.25
    duration = 50.0 * 365.25
    pad_around = 40
    t_step = 0.2

    site = Site("LSST")
    observer = Observer(
        longitude=site.longitude * u.deg,
        latitude=site.latitude * u.deg,
        elevation=site.height * u.m,
        name="LSST",
    )

    results = []

    mjd = mjd_start + 0
    counter = 0
    names = [
        "night",
        "sunset",
        "sun_n12_setting",
        "sun_n18_setting",
        "sun_n18_rising",
        "sun_n12_rising",
        "sunrise",
        "moonrise",
        "moonset",
    ]
    types = [int]
    types.extend([float] * (len(names) - 1))
    base_al = np.zeros(1, dtype=list(zip(names, types)))

    while mjd < (mjd_start + duration):
        times = Time(mjd, format="mjd")
        sunsets = observer.sun_set_time(times, which="next")
        times = sunsets

        almanac = base_al.copy()
        almanac["sunset"] = sunsets.mjd

        almanac["moonset"] = observer.moon_set_time(times, which="next").mjd
        almanac["moonrise"] = observer.moon_rise_time(times, which="next").mjd
        almanac["sun_n12_setting"] = observer.twilight_evening_nautical(times, which="next").mjd
        times = observer.twilight_evening_astronomical(times, which="next")
        almanac["sun_n18_setting"] = times.mjd
        almanac["sun_n18_rising"] = observer.twilight_morning_astronomical(times, which="next").mjd
        almanac["sun_n12_rising"] = observer.twilight_morning_nautical(times, which="next").mjd
        almanac["sunrise"] = observer.sun_rise_time(times, which="next").mjd
        results.append(almanac)
        mjd = almanac["sunrise"] + t_step

        progress = (mjd - mjd_start) / duration * 100
        text = "\rprogress = %.2f%%" % progress
        sys.stdout.write(text)
        sys.stdout.flush()

    almanac = np.concatenate(results)
    almanac["night"] = np.arange(almanac["night"].size)

    np.savez("sunsets.npz", almanac=almanac)

    # runs in real    193m21.246s



__all__ = ("ScheduledDowntimeData",)

import os
import sqlite3

import numpy as np
from astropy.time import Time, TimeDelta

from rubin_scheduler.data import get_data_dir


class ScheduledDowntimeData:
    """Read the scheduled downtime data.

    This class deals with the scheduled downtime information that
    was previously produced for OpSim version 3.

    Parameters
    ----------
    start_time : `astropy.time.Time`
        The time of the start of the simulation.
        The cloud database will be assumed to start on Jan 01 of the
        same year.
    cloud_db : `str`, optional
        The full path name for the cloud database. Default None,
        which will use the database stored in the module
        ($SIMS_CLOUDMODEL_DIR/data/cloud.db).
    start_of_night_offset : `float`, optional
        The fraction of a day to offset from MJD.0 to reach the defined
        start of a night ('noon' works). Default 0.16 (UTC midnight
        in Chile) - 0.5 (minus half a day) = -0.34
    """

    def __init__(self, start_time, scheduled_downtime_db=None, start_of_night_offset=-0.34):
        self.scheduled_downtime_db = scheduled_downtime_db
        if self.scheduled_downtime_db is None:
            self.scheduled_downtime_db = os.path.join(get_data_dir(), "site_models", "scheduled_downtime.db")

        # downtime database starts in Jan 01 of the year of the
        # start of the simulation.
        year_start = start_time.datetime.year
        self.night0 = Time("%d-01-01" % year_start, format="isot", scale="tai") + TimeDelta(
            start_of_night_offset, format="jd"
        )

        # Scheduled downtime data is a np.ndarray of start / end /
        # activity for each scheduled downtime.
        self.downtime = None
        self.read_data()

    def __call__(self):
        """Return the current (if any) and any future scheduled downtimes.

        Parameters
        ----------
        time : `astropy.time.Time`
            Time in the simulation for which to find the current downtime.

        Returns
        -------
        downtime : `np.ndarray`
            The array of all unscheduled downtimes, with keys for
            'start', 'end', 'activity', corresponding to
            astropy.time.Time, astropy.time.Time, and str.
        """
        return self.downtime

    def _downtime_status(self, time):
        """Look behind the scenes at the downtime status/next values"""
        next_start = self.downtime["start"].searchsorted(time, side="right")
        next_end = self.downtime["end"].searchsorted(time, side="right")
        if next_start > next_end:
            current = self.downtime[next_end]
        else:
            current = None
        future = self.downtime[next_start:]
        return current, future

    def read_data(self):
        """Read the scheduled downtime information from disk and
        translate to astropy.time.Times.

        This function gets the appropriate database file and creates
        the set of scheduled downtimes from it. The default behavior
        is to use the module stored database. However, an alternate
        database file can be provided. The alternate database file needs
        to have a table called *Downtime* with the following columns:

        night : `int`
            The night (from start of simulation) the downtime occurs.
        duration : `int`
            The duration (units=days) of the downtime.
        activity : `str`
            A description of the activity involved.
        """
        # Read from database.
        starts = []
        ends = []
        acts = []
        with sqlite3.connect(self.scheduled_downtime_db) as conn:
            cur = conn.cursor()
            cur.execute("select * from Downtime;")
            for row in cur:
                start_night = int(row[0])
                start_night = self.night0 + TimeDelta(start_night, format="jd")
                n_down = int(row[1])
                end_night = start_night + TimeDelta(n_down, format="jd")
                activity = row[2]
                starts.append(start_night)
                ends.append(end_night)
                acts.append(activity)
            cur.close()
        self.downtime = np.array(
            list(zip(starts, ends, acts)),
            dtype=[("start", "O"), ("end", "O"), ("activity", "O")],
        )

    def total_downtime(self):
        """Return total downtime (in days).

        Returns
        -------
        total : `int`
            Total number of downtime days.
        """
        total = 0
        for td in self.downtime["end"] - self.downtime["start"]:
            total += td.jd
        return total



import os
import unittest

import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler.example import example_scheduler, run_sched
from rubin_scheduler.utils import survey_start_mjd


class TestExample(unittest.TestCase):

    @unittest.skipUnless(
        os.path.isfile(os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")),
        "Test data not available.",
    )
    def test_example(self):
        """Test the example scheduler executes all the expected surveys"""
        mjd_start = survey_start_mjd()
        scheduler = example_scheduler(mjd_start=mjd_start)
        observatory, scheduler, observations = run_sched(scheduler, mjd_start=mjd_start, survey_length=7)
        u_notes = np.unique(observations["scheduler_note"])

        # Note that some of these may change and need to be updated if
        # survey start date changes, e.g., different DDFs in season,
        # or different lunar phase means different filters get picked
        # for the blobs
        notes_to_check = [
            "DD:COSMOS",
            "blob_long, gr, a",
            "blob_long, gr, b",
            "greedy",
            "long",
            "pair_15, iz, a",
            "pair_15, iz, b",
            "pair_15, ri, a",
            "pair_15, ri, b",
            "pair_15, yy, a",
            "pair_15, yy, b",
            "pair_33, gr, a",
            "pair_33, gr, b",
            "pair_33, ri, a",
            "pair_33, ug, a",
            "pair_33, ug, b",
            "pair_33, yy, a",
            "pair_33, yy, b",
            "pair_33, zy, a",
            "pair_33, zy, b",
            "twilight_near_sun, 0",
            "twilight_near_sun, 1",
            "twilight_near_sun, 2",
            "twilight_near_sun, 3",
        ]

        for note in notes_to_check:
            assert note in u_notes

        for note in u_notes:
            # If this fails, time to add something to notes_to_check
            assert note in u_notes


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from rubin_scheduler.site_models import SeeingModel


class TestSeeingModel(unittest.TestCase):
    def test_fwhm_system_zenith(self):
        # Check calculation is being done as expected.
        seeing_model = SeeingModel()
        self.assertAlmostEqual(seeing_model.fwhm_system_zenith, 0.39862262855989494, places=7)

    def test_fwhm_geom_eff(self):
        # Check that the translation between FWHM effective and
        # geometric is done as expected.
        # (note that fwhmEff_tofwhmGeom & fwhmGeom_to_fwhmEff are
        # static methods) Document-20160 for reference.
        fwhm_eff = 1.23
        fwhm_geom = 0.822 * fwhm_eff + 0.052
        self.assertEqual(fwhm_geom, SeeingModel.fwhm_eff_to_fwhm_geom(fwhm_eff))
        self.assertEqual(fwhm_eff, SeeingModel.fwhm_geom_to_fwhm_eff(fwhm_geom))

    def test_call(self):
        # Check the calculation from fwhm_500 to fwhm_eff/fwhm_geom.
        # Use simple effective wavelengths and airmass values.
        filter_list = ["500", "1000"]
        effwavelens = np.array([500.0, 1000.0])
        seeing_model = SeeingModel(filter_list=filter_list, eff_wavelens=effwavelens)
        # Simple fwhm_500 input.
        fwhm_500 = 1.0
        # Single airmass.
        airmass = 1.0
        seeing = seeing_model(fwhm_500, airmass)
        fwhm_eff = seeing["fwhmEff"]
        # Check shape of returned values.
        self.assertEqual(fwhm_eff.shape, (len(seeing_model.eff_wavelens),))
        # Check actual value of seeing in @ wavelen[0] @ zenith
        # after addition of system.
        fwhm_system = seeing_model.fwhm_system_zenith
        expected_fwhm_eff = 1.16 * np.sqrt(fwhm_system**2 + 1.04 * fwhm_500**2)
        self.assertAlmostEqual(fwhm_eff[0], expected_fwhm_eff, 15)
        # Check expected value if we remove the system component.
        seeing_model.fwhm_system_zenith = 0
        seeing = seeing_model(fwhm_500, airmass)
        expected_fwhm_eff = 1.16 * np.sqrt(1.04) * fwhm_500
        self.assertAlmostEqual(seeing["fwhmEff"][0], expected_fwhm_eff, 15)
        # Check scaling with wavelength (remove system component).
        expected_fwhm_eff = 1.16 * np.sqrt(1.04) * fwhm_500 * np.power(500.0 / effwavelens[1], 0.3)
        self.assertAlmostEqual(seeing["fwhmEff"][1], expected_fwhm_eff, places=15)
        # Multiple airmasses.
        airmass = np.array([1.0, 1.5])
        seeing = seeing_model(fwhm_500, airmass)
        self.assertEqual(seeing["fwhmEff"].shape, (len(seeing_model.eff_wavelens), len(airmass)))
        expected_fwhm_eff = fwhm_500 * 1.16 * np.sqrt(1.04)
        self.assertEqual(seeing["fwhmEff"][0][0], expected_fwhm_eff)
        # Check scaling with airmass.
        expected_fwhm_eff = expected_fwhm_eff * np.power(airmass[1], 0.6)
        self.assertAlmostEqual(seeing["fwhmEff"][0][1], expected_fwhm_eff, places=15)


if __name__ == "__main__":
    unittest.main()



"""Footprints: Take sky area maps and turn them into dynamic `footprint`
objects which understand seasons and time, in order to weight area on sky
appropriately for a given time.
"""

__all__ = (
    "ra_dec_hp_map",
    "calc_norm_factor",
    "calc_norm_factor_array",
    "StepLine",
    "Footprints",
    "Footprint",
    "StepSlopes",
    "ConstantFootprint",
    "BasePixelEvolution",
    "slice_wfd_area_quad",
    "slice_wfd_indx",
    "slice_quad_galactic_cut",
    "make_rolling_footprints",
)

import healpy as hp
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

from rubin_scheduler.utils import _hpid2_ra_dec

from .utils import set_default_nside


def make_rolling_footprints(
    fp_hp=None,
    mjd_start=60218.0,
    sun_ra_start=3.27717639,
    nslice=2,
    scale=0.8,
    nside=32,
    wfd_indx=None,
    order_roll=0,
    n_cycles=None,
    n_constant_start=2,
    n_constant_end=6,
    verbose=False,
    uniform=True,
):
    """
    Generate rolling footprints

    Parameters
    ----------
    fp_hp : dict-like
        A dict with filtername keys and HEALpix map values
    mjd_start : `float`
        The starting date of the survey.
    sun_ra_start : `float`
        The RA of the sun at the start of the survey
    nslice : `int`
        How much to slice the sky up. Can be 2, 3, 4, or 6.
    scale : `float`
        The strength of the rolling, value of 1 is full power rolling.
        Zero is no rolling.
    wfd_indx : array of ints
        The indices of the HEALpix map that are to be included in the rolling.
    order_roll : `int`
        Change the order of when bands roll. Default 0.
    n_cycles : `int`
        Number of complete rolling cycles to attempt. If None, defaults to 3
        full cycles for nslice=2, 2 cycles for nslice=3 or 4, and 1 cycle for
        nslice=6.
    n_constant_start : `int`
        The number of constant non-rolling seasons to start with. Anything
        less than 2 will start rolling too early near Y1. Defaults to 2.
    n_constant_end : `int`
        The number of constant seasons to end the survey with. Defaults to 6.

    Returns
    -------
    Footprints object
    """

    nc_default = {2: 3, 3: 2, 4: 2, 6: 1}
    if n_cycles is None:
        n_cycles = nc_default[nslice]

    hp_footprints = fp_hp

    D = 1.0 - scale
    U = nslice - D * (nslice - 1)

    start = [1.0] * n_constant_start
    # After n_cycles, just go to no-rolling for 6 years.
    end = [1.0] * n_constant_end

    rolling = [U] + [D] * (nslice - 1)
    rolling = np.roll(rolling, order_roll).tolist()

    all_slopes = []
    if uniform:
        extra_constant = [1]
    else:
        extra_constant = []

    for i in range(nslice):
        _roll = np.roll(rolling, i).tolist() + extra_constant
        all_slopes.append(start + _roll * n_cycles + end)
    for i in range(nslice):
        _roll = np.roll(rolling, i).tolist() + extra_constant
        _roll = [_roll[-1]] + _roll[1:-1] + [_roll[0]]
        all_slopes.append(start + _roll * n_cycles + end)
    dvals = {
        1: "1",
        D: "D",
        U: "U",
    }

    abc = ["a", "b", "c", "d", "e", "f", "g", "h"]
    slice_names = ["slice %s" % abc[i] for i in range(nslice)]
    for i, s in enumerate(all_slopes):
        if i >= nslice:
            sname = slice_names[i - nslice] + " w/ ra - sun_ra in [90, 270]"
        else:
            sname = slice_names[i] + " w/ ra - sun_ra in [270, 90]"
        if verbose:
            print(sname + ": " + " ".join([dvals[x] for x in s]))

    fp_non_wfd = Footprint(mjd_start, sun_ra_start=sun_ra_start, nside=nside)
    rolling_footprints = []
    for i in range(len(all_slopes)):
        step_func = StepSlopes(rise=all_slopes[i])
        rolling_footprints.append(
            Footprint(
                mjd_start,
                sun_ra_start=sun_ra_start,
                step_func=step_func,
                nside=nside,
            )
        )

    wfd = hp_footprints["r"] * 0
    if wfd_indx is None:
        wfd_indx = np.where(hp_footprints["r"] == 1)[0]

    wfd[wfd_indx] = 1
    non_wfd_indx = np.where(wfd == 0)[0]

    if uniform:
        split_wfd_indices = slice_quad_galactic_cut(
            hp_footprints,
            nslice=nslice,
            wfd_indx=wfd_indx,
            ra_range=(sun_ra_start + 1.5 * np.pi, sun_ra_start + np.pi / 2),
        )

        split_wfd_indices_delayed = slice_quad_galactic_cut(
            hp_footprints,
            nslice=nslice,
            wfd_indx=wfd_indx,
            ra_range=(sun_ra_start + np.pi / 2, sun_ra_start + 1.5 * np.pi),
        )
    else:
        split_wfd_indices = slice_quad_galactic_cut(hp_footprints, nslice=nslice, wfd_indx=wfd_indx)

    for key in hp_footprints:
        temp = hp_footprints[key] + 0
        temp[wfd_indx] = 0
        fp_non_wfd.set_footprint(key, temp)

        for i in range(nslice):
            # make a copy of the current filter
            temp = hp_footprints[key] + 0
            # Set the non-rolling area to zero
            temp[non_wfd_indx] = 0

            indx = split_wfd_indices[i]
            # invert the indices
            ze = temp * 0
            ze[indx] = 1
            temp = temp * ze
            rolling_footprints[i].set_footprint(key, temp)
        if uniform:
            for _i in range(nslice, nslice * 2):
                # make a copy of the current filter
                temp = hp_footprints[key] + 0
                # Set the non-rolling area to zero
                temp[non_wfd_indx] = 0

                indx = split_wfd_indices_delayed[_i - nslice]
                # invert the indices
                ze = temp * 0
                ze[indx] = 1
                temp = temp * ze
                rolling_footprints[_i].set_footprint(key, temp)

    result = Footprints([fp_non_wfd] + rolling_footprints)
    return result


def _is_in_ra_range(ra, low, high):
    _low = low % (2.0 * np.pi)
    _high = high % (2.0 * np.pi)
    if _low <= _high:
        return (ra >= _low) & (ra <= _high)
    else:
        return (ra >= _low) | (ra <= _high)


def slice_quad_galactic_cut(target_map, nslice=2, wfd_indx=None, ra_range=None):
    """
    Helper function for generating rolling footprints

    Parameters
    ----------
    target_map : dict of HEALpix maps
        The final desired footprint as HEALpix maps. Keys are filter names
    nslice : `int`
        The number of slices to make, can be 2 or 3.
    wfd_indx : array of ints
        The indices of target_map that should be used for rolling.
        If None, assumes the rolling area should be where target_map['r'] == 1.
    ra_range : tuple of floats, optional
        If not None, then the indices are restricted to the given RA range
        in radians.
    """

    ra, dec = ra_dec_hp_map(nside=hp.npix2nside(target_map["r"].size))

    coord = SkyCoord(ra=ra * u.rad, dec=dec * u.rad)
    _, gal_lat = coord.galactic.l.deg, coord.galactic.b.deg

    indx_north = np.intersect1d(np.where(gal_lat >= 0)[0], wfd_indx)
    indx_south = np.intersect1d(np.where(gal_lat < 0)[0], wfd_indx)

    splits_north = slice_wfd_area_quad(target_map, nslice=nslice, wfd_indx=indx_north)
    splits_south = slice_wfd_area_quad(target_map, nslice=nslice, wfd_indx=indx_south)

    slice_indx = []
    for j in np.arange(nslice):
        indx_temp = []
        for i in np.arange(j + 1, nslice * 2 + 1, nslice):
            indx_temp += indx_north[splits_north[i - 1] : splits_north[i]].tolist()
            indx_temp += indx_south[splits_south[i - 1] : splits_south[i]].tolist()
        slice_indx.append(indx_temp)

    if ra_range is not None:
        ra_indx = np.where(_is_in_ra_range(ra, *ra_range))[0]
        for j in range(nslice):
            slice_indx[j] = np.intersect1d(ra_indx, slice_indx[j])

    return slice_indx


def slice_wfd_area_quad(target_map, nslice=2, wfd_indx=None):
    """
    Divide a healpix map in an intelligent way

    Parameters
    ----------
    target_map : dict of HEALpix arrays
        The input map to slice
    nslice : int
        The number of slices to divide the sky into (gets doubled).
    wfd_indx : array of int
        The indices of the healpix map to consider as part of the WFD area
        that will be split.
        If set to None, the pixels where target_map['r'] == 1 are
        considered as WFD.
    """
    nslice2 = nslice * 2

    wfd = target_map["r"] * 0
    if wfd_indx is None:
        wfd_indices = np.where(target_map["r"] == 1)[0]
    else:
        wfd_indices = wfd_indx
    wfd[wfd_indices] = 1
    wfd_accum = np.cumsum(wfd)
    split_wfd_indices = np.floor(np.max(wfd_accum) / nslice2 * (np.arange(nslice2) + 1)).astype(int)
    split_wfd_indices = split_wfd_indices.tolist()
    split_wfd_indices = [0] + split_wfd_indices

    return split_wfd_indices


def ra_dec_hp_map(nside=None):
    """
    Return all the RA,dec points for the centers of a healpix map, in radians.
    """
    if nside is None:
        nside = set_default_nside()
    ra, dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
    return ra, dec


def slice_wfd_indx(target_map, nslice=2, wfd_indx=None):
    """
    simple map split
    """

    wfd = target_map["r"] * 0
    if wfd_indx is None:
        wfd_indx = np.where(target_map["r"] == 1)[0]
    wfd[wfd_indx] = 1
    wfd_accum = np.cumsum(wfd)
    split_wfd_indices = np.floor(np.max(wfd_accum) / nslice * (np.arange(nslice) + 1)).astype(int)
    split_wfd_indices = split_wfd_indices.tolist()
    split_wfd_indices = [0] + split_wfd_indices

    return split_wfd_indices


def _is_in_ra_range(ra, low, high):
    _low = low % (2.0 * np.pi)
    _high = high % (2.0 * np.pi)
    if _low <= _high:
        return (ra >= _low) & (ra <= _high)
    else:
        return (ra >= _low) | (ra <= _high)


class BasePixelEvolution:
    """Helper class that can be used to describe the time evolution of a
    HEALpix in a footprint.
    """

    def __init__(self, period=365.25, rise=1.0, t_start=0.0):
        self.period = period
        self.rise = rise
        self.t_start = t_start

    def __call__(self, mjd_in, phase):
        pass


class StepLine(BasePixelEvolution):

    def __call__(self, mjd_in, phase):
        t = mjd_in + phase - self.t_start
        n_periods = np.floor(t / (self.period))
        result = n_periods * self.rise
        tphased = t % self.period
        step_area = np.where(tphased > self.period / 2.0)[0]
        result[step_area] += (tphased[step_area] - self.period / 2) * self.rise / (0.5 * self.period)
        result[np.where(t < 0)] = 0
        return result


class StepSlopes(BasePixelEvolution):
    def __call__(self, mjd_in, phase):
        steps = np.array(self.rise)
        t = mjd_in + phase - self.t_start
        season = np.floor(t / (self.period))
        season = season.astype(int)
        plateus = np.cumsum(steps) - steps[0]
        result = plateus[season]
        tphased = t % self.period
        step_area = np.where(tphased > self.period / 2.0)[0]
        result[step_area] += (
            (tphased[step_area] - self.period / 2) * steps[season + 1][step_area] / (0.5 * self.period)
        )
        result[np.where(t < 0)] = 0

        return result


class Footprint:
    """An object to compute the desired survey footprint at a given time

    Parameters
    ----------
    mjd_start : float
        The MJD the survey starts on
    sun_ra_start : float
        The RA of the sun at the start of the survey (radians)

    """

    def __init__(
        self,
        mjd_start,
        sun_ra_start=0,
        nside=32,
        filters=None,
        period=365.25,
        step_func=None,
    ):
        if filters is None:
            filters = {"u": 0, "g": 1, "r": 2, "i": 3, "z": 4, "y": 5}
        self.period = period
        self.nside = nside
        if step_func is None:
            step_func = StepLine()
        self.step_func = step_func
        self.mjd_start = mjd_start
        self.sun_ra_start = sun_ra_start
        self.npix = hp.nside2npix(nside)
        self.filters = filters
        self.ra, self.dec = _hpid2_ra_dec(self.nside, np.arange(self.npix))
        # Set the phase of each healpixel.
        # If RA to sun is zero, we are at phase np.pi/2.
        # This is similar to the season calculation, except
        # that phase and season are offset by 90 degrees.
        self.phase = (-self.ra + self.sun_ra_start + np.pi / 2) % (2.0 * np.pi)
        self.phase = self.phase * (self.period / 2.0 / np.pi)
        # Empty footprints to start
        self.out_dtype = list(zip(filters, [float] * len(filters)))
        self.footprints = np.zeros((len(filters), self.npix), dtype=float)
        self.estimate = np.zeros((len(filters), self.npix), dtype=float)
        self.current_footprints = np.zeros((len(filters), self.npix), dtype=float)
        self.zero = self.step_func(0.0, self.phase)
        self.mjd_current = None

    def set_footprint(self, filtername, values):
        self.footprints[self.filters[filtername], :] = values

    def get_footprint(self, filtername):
        return self.footprints[self.filters[filtername], :]

    def _update_mjd(self, mjd, norm=True):
        if mjd != self.mjd_current:
            self.mjd_current = mjd
            t_elapsed = mjd - self.mjd_start

            norm_coverage = self.step_func(t_elapsed, self.phase)
            norm_coverage -= self.zero
            self.current_footprints = self.footprints * norm_coverage
            c_sum = np.nansum(self.current_footprints)
            if norm:
                if c_sum != 0:
                    self.current_footprints = self.current_footprints / c_sum

    def arr2struc(self, inarr):
        """Take an array and convert it to labeled struc array"""
        result = np.empty(self.npix, dtype=self.out_dtype)
        for key in self.filters:
            result[key] = inarr[self.filters[key]]
        # Argle bargel, why doesn't this view work?
        # struc = inarr.view(dtype=self.out_dtype).squeeze()
        return result

    def estimate_counts(self, mjd, nvisits=2.2e6, fov_area=9.6):
        """Estimate the counts we'll get after some time and visits"""
        pix_area = hp.nside2pixarea(self.nside, degrees=True)
        pix_per_visit = fov_area / pix_area
        self._update_mjd(mjd, norm=True)
        self.estimate = self.current_footprints * pix_per_visit * nvisits
        return self.arr2struc(self.estimate)

    def __call__(self, mjd, norm=True):
        """
        Parameters
        ----------
        mjd : `float`
            Current MJD.
        norm : `bool`
            If normalized, the footprint retains the same range of values
            over time.

        Returns
        -------
        current_footprints : `np.ndarray`, (6, N)
            A numpy structured array with the updated normalized number of
            observations that should be requested at each Healpix.
            Multiply by the number of HEALpix observations (all filter), to
            convert to the number of observations desired.
        """
        self._update_mjd(mjd, norm=norm)
        return self.arr2struc(self.current_footprints)


class ConstantFootprint(Footprint):
    def __init__(self, nside=32, filters=None):
        if filters is None:
            filters = {"u": 0, "g": 1, "r": 2, "i": 3, "z": 4, "y": 5}
        self.nside = nside
        self.filters = filters
        self.npix = hp.nside2npix(nside)
        self.footprints = np.zeros((len(filters), self.npix), dtype=float)
        self.out_dtype = list(zip(filters, [float] * len(filters)))
        self.to_return = self.arr2struc(self.footprints)

    def set_footprint(self, filtername, values):
        self.footprints[self.filters[filtername], :] = values
        self.to_return = self.arr2struc(self.footprints)

    def __call__(self, mjd, array=False):
        return self.to_return


class Footprints(Footprint):
    """An object to combine multiple Footprint objects."""

    def __init__(self, footprint_list):
        self.footprint_list = footprint_list
        self.mjd_current = None
        self.current_footprints = 0
        # Should probably run a check that all the footprints are compatible
        # (same nside, etc)
        self.npix = footprint_list[0].npix
        self.out_dtype = footprint_list[0].out_dtype
        self.filters = footprint_list[0].filters
        self.nside = footprint_list[0].nside

        self.footprints = np.zeros((len(self.filters), self.npix), dtype=float)
        for fp in self.footprint_list:
            self.footprints += fp.footprints

    def set_footprint(self, filtername, values):
        pass

    def _update_mjd(self, mjd, norm=True):
        if mjd != self.mjd_current:
            self.mjd_current = mjd
            self.current_footprints = 0.0
            for fp in self.footprint_list:
                fp._update_mjd(mjd, norm=False)
                self.current_footprints += fp.current_footprints
            c_sum = np.sum(self.current_footprints)
            if norm:
                if c_sum != 0:
                    self.current_footprints = self.current_footprints / c_sum


def calc_norm_factor(goal_dict, radius=1.75):
    """Calculate how to normalize a Target_map_basis_function.
    This is basically:
    the area of the fov / area of a healpixel  /
    the sum of all of the weighted-healpix values in the footprint.

    Parameters
    -----------
    goal_dict : dict of healpy maps
        The target goal map(s) being used
    radius : float (1.75)
        Radius of the FoV (degrees)

    Returns
    -------
    Value to use as Target_map_basis_function norm_factor kwarg
    """
    all_maps_sum = 0
    for key in goal_dict:
        good = np.where(goal_dict[key] > 0)
        all_maps_sum += goal_dict[key][good].sum()
    nside = hp.npix2nside(goal_dict[key].size)
    hp_area = hp.nside2pixarea(nside, degrees=True)
    norm_val = radius**2 * np.pi / hp_area / all_maps_sum
    return norm_val


def calc_norm_factor_array(goal_map, radius=1.75):
    """Calculate how to normalize a Target_map_basis_function.
    This is basically:
    the area of the fov / area of a healpixel  /
    the sum of all of the weighted-healpix values in the footprint.

    Parameters
    -----------
    goal_map : recarray of healpy maps
        The target goal map(s) being used
    radius : float
        Radius of the FoV (degrees)

    Returns
    -------
    Value to use as Target_map_basis_function norm_factor kwarg
    """
    all_maps_sum = 0
    for key in goal_map.dtype.names:
        good = np.where(goal_map[key] > 0)
        all_maps_sum += goal_map[key][good].sum()
    nside = hp.npix2nside(goal_map[key].size)
    hp_area = hp.nside2pixarea(nside, degrees=True)
    norm_val = radius**2 * np.pi / hp_area / all_maps_sum
    return norm_val



__all__ = ("ToOScriptedSurvey", "gen_too_surveys")

from copy import copy

import numpy as np

from rubin_scheduler.scheduler.surveys import BaseMarkovSurvey, ScriptedSurvey
from rubin_scheduler.scheduler.utils import (
    ScheduledObservationArray,
    comcam_tessellate,
    order_observations,
    thetaphi2xyz,
    xyz2thetaphi,
)
from rubin_scheduler.site_models import _read_fields
from rubin_scheduler.utils import _approx_ra_dec2_alt_az, _ra_dec2_hpid


def rotx(theta, x, y, z):
    """rotate the x,y,z points theta radians about x axis"""
    sin_t = np.sin(theta)
    cos_t = np.cos(theta)
    xp = x
    yp = y * cos_t + z * sin_t
    zp = -y * sin_t + z * cos_t
    return xp, yp, zp


DEFAULT_EXP_TIME = 29.2


class ToOScriptedSurvey(ScriptedSurvey, BaseMarkovSurvey):
    """If there is a new ToO event, generate a set of scripted
    observations to try and follow it up.

    Parameters
    ----------
    times : list of floats
        The times after the detection that observations
        should be attempted (hours). Default [1, 2, 4, 24, 48]
    filters_at_times : list of str
        The filters that should be observed at each time in `times`.
        Default ["gz", "gz", "gz", "gz", "gz", "gz"]
    nvis : list of int
        The number of visits per filter at each time in `times`.
        Default [1, 1, 1, 1, 6, 6]
    exptimes : list of floats
        The exposure times to use for each time in `times`.
        Default [DEFAULT_EXP_TIME]*5,
    alt_min : float
        Do not attempt observations below this limit (degrees).
        Note the telescope alt limit is 20 degrees, however, slew
        and filter change time means alt_min here should be set higher
        (otherwise, target will pass altitude check, but then fail
        to observe by the time the telescope gets there).

    """

    def __init__(
        self,
        basis_functions,
        followup_footprint=None,
        nside=32,
        reward_val=1e6,
        times=[1, 2, 4, 24, 48],
        filters_at_times=["gz", "gz", "gz", "gz", "gz", "gz"],
        nvis=[1, 1, 1, 1, 6, 6],
        exptimes=[DEFAULT_EXP_TIME] * 6,
        camera="LSST",
        survey_name="ToO",
        flushtime=2.0,
        mjd_tol=1.0 / 24.0,
        dist_tol=0.5,
        alt_min=25.0,
        alt_max=85.0,
        HA_min=5,
        HA_max=19,
        ignore_obs=None,
        dither=True,
        seed=42,
        npositions=20000,
        n_snaps=2,
        n_usnaps=1,
        id_start=1,
        detailers=None,
        too_types_to_follow=[""],
        split_long=False,
    ):

        # Make sure lists all match
        check = np.unique([len(filters_at_times), len(times), len(nvis), len(exptimes)])
        if np.size(check) > 1:
            raise ValueError("lengths of times, filters_at_times, nvis, and exptimes must match.")

        # Figure out what else I need to super here

        if ignore_obs is None:
            ignore_obs = []

        self.basis_functions = basis_functions
        self.survey_name = survey_name
        self.followup_footprint = followup_footprint
        self.last_event_id = -1
        self.night = -1
        self.reward_val = reward_val
        self.times = np.array(times) / 24.0  # to days
        self.filters_at_times = filters_at_times
        self.exptimes = exptimes
        self.nvis = nvis
        self.n_snaps = n_snaps
        self.n_usnaps = n_usnaps
        self.nside = nside
        self.flushtime = flushtime / 24.0
        self.mjd_tol = mjd_tol
        self.dist_tol = np.radians(dist_tol)
        self.alt_min = np.radians(alt_min)
        self.alt_max = np.radians(alt_max)
        self.HA_min = HA_min
        self.HA_max = HA_max
        self.ignore_obs = ignore_obs
        self.extra_features = {}
        self.extra_basis_functions = {}
        self.detailers = []
        self.dither = dither
        self.id_start = id_start
        self.detailers = detailers
        self.last_mjd = -1
        self.too_types_to_follow = too_types_to_follow
        self.split_long = split_long

        self.camera = camera
        # Load the OpSim field tesselation and map healpix to fields
        if self.camera == "LSST":
            ra, dec = _read_fields()
            self.fields_init = np.empty(ra.size, dtype=list(zip(["RA", "dec"], [float, float])))
            self.fields_init["RA"] = ra
            self.fields_init["dec"] = dec
        elif self.camera == "comcam":
            self.fields_init = comcam_tessellate()
        else:
            ValueError('camera %s unknown, should be "LSST" or "comcam"' % camera)
        self.fields = self.fields_init.copy()

        self.hp2fields = np.array([])
        self._hp2fieldsetup(self.fields["RA"], self.fields["dec"])

        # Don't bother with checking if we can run before twilight ends
        self.before_twi_check = False

        # Initialize the list of scripted observations
        self.clear_script()

        # Generate and store rotation positions to use.
        # This way, if different survey objects are seeded the same, they will
        # use the same dither positions each night
        rng = np.random.default_rng(seed)
        self.lon = rng.random(npositions) * np.pi * 2
        # Make sure latitude points spread correctly
        # http://mathworld.wolfram.com/SpherePointPicking.html
        self.lat = np.arccos(2.0 * rng.random(npositions) - 1.0)
        self.lon2 = rng.random(npositions) * np.pi * 2
        self.spin_indx = 0

        # list to keep track of alerts we have already seen
        self.seen_alerts = []

    # Need to make sure new set_script call doesn't clobber old script!
    def set_script(self, obs_wanted, append=True):
        """
        Parameters
        ----------
        obs_wanted : rubin_scheduler.scheduler.utils.ScheduledObservationArray
            The observations that should be executed.
        append : bool
            Should the obs_wanted be appended to any script already set?
        """

        obs_wanted.sort(order=["mjd", "filter"])
        # Give each desired observation a unique "scripted ID". To be used for
        # matching and logging later.
        obs_wanted["scripted_id"] = np.arange(self.id_start, self.id_start + np.size(obs_wanted))
        # Update so if we set the script again the IDs will not be reused.
        self.id_start = np.max(obs_wanted["scripted_id"]) + 1

        # If we already have a script and append
        if append & (self.obs_wanted is not None):
            self.obs_wanted = np.concatenate([self.obs_wanted, obs_wanted])
            self.obs_wanted.sort(order=["mjd", "filter"])
        else:
            self.obs_wanted = obs_wanted

        self.mjd_start = self.obs_wanted["mjd"] - self.obs_wanted["mjd_tol"]
        # Here is the atribute that core scheduler checks to
        # broadcast scheduled observations in the conditions object.
        self.scheduled_obs = self.obs_wanted["mjd"]

    def _check_list(self, conditions):
        """Check to see if the current mjd is good"""
        observation = None
        if self.obs_wanted is not None:
            # Scheduled observations that are in the right
            # time window and have not been executed
            in_time_window = np.where(
                (self.mjd_start < conditions.mjd)
                & (self.obs_wanted["flush_by_mjd"] > conditions.mjd)
                & (~self.obs_wanted["observed"])
            )[0]

            if np.size(in_time_window) > 0:
                pass_checks = self._check_alts_ha(self.obs_wanted[in_time_window], conditions)
                matches = in_time_window[pass_checks]
            else:
                matches = []

            if np.size(matches) > 0:
                observation = self._slice2obs(self.obs_wanted[matches[0]])

        return observation

    def flush_script(self, conditions):
        """Remove things from the script that aren't needed anymore"""
        if self.obs_wanted is not None:
            still_relevant = np.where(
                (self.obs_wanted["observed"] == False) & (self.obs_wanted["flush_by_mjd"] > conditions.mjd)
            )[0]

            if np.size(still_relevant) > 0:
                observations = self.obs_wanted[still_relevant]
                self.set_script(observations, append=False)
            else:
                self.clear_script()

    def _spin_fields(self):
        """Spin the field tessellation to generate a random orientation

        The default field tesselation is rotated randomly in longitude,
        and then the pole is rotated to a random point on the sphere.

        Automatically advances self.spin_indx when called.

        """
        lon = self.lon[self.spin_indx]
        lat = self.lat[self.spin_indx]
        lon2 = self.lon2[self.spin_indx]

        # rotate longitude
        ra = (self.fields_init["RA"] + lon) % (2.0 * np.pi)
        dec = copy(self.fields_init["dec"])

        # Now to rotate ra and dec about the x-axis
        x, y, z = thetaphi2xyz(ra, dec + np.pi / 2.0)
        xp, yp, zp = rotx(lat, x, y, z)
        theta, phi = xyz2thetaphi(xp, yp, zp)
        dec = phi - np.pi / 2
        ra = theta + np.pi

        # One more RA rotation
        ra = (ra + lon2) % (2.0 * np.pi)

        self.fields["RA"] = ra
        self.fields["dec"] = dec
        # Rebuild the kdtree with the new positions
        # XXX-may be doing some ra,dec to conversions xyz more
        # than needed.
        self._hp2fieldsetup(ra, dec)

        # Advance the spin index
        self.spin_indx += 1

    def _tesselate(self, hpid_to_observe):
        self._spin_fields()
        field_ids = np.unique(self.hp2fields[hpid_to_observe])
        # Put the fields in a good order.
        better_order = order_observations(self.fields["RA"][field_ids], self.fields["dec"][field_ids])
        ras = self.fields["RA"][field_ids[better_order]]
        decs = self.fields["dec"][field_ids[better_order]]

        return ras, decs

    def _new_event(self, target_o_o, conditions):
        """A new ToO event, generate any observations for followup"""
        # flush out any old observations or ones that have been completed
        self.flush_script(conditions)

        # Check that this is the type of ToO we are supposed to trigger on
        correct_type = False
        for type_to_follow in self.too_types_to_follow:
            if type_to_follow in target_o_o.too_type:
                correct_type = True

        # Check that we have not seen this alert yet
        unseen = False
        if target_o_o.id not in self.seen_alerts:
            unseen = True

        if correct_type & unseen:
            self.seen_alerts.append(target_o_o.id)
            # Check that the event center is in the footprint
            # we want to observe
            hpid_center = _ra_dec2_hpid(self.nside, target_o_o.ra_rad_center, target_o_o.dec_rad_center)
            if self.followup_footprint[hpid_center] > 0:
                target_area = self.followup_footprint * target_o_o.footprint
                # generate a list of pointings for that area
                hpid_to_observe = np.where(target_area > 0)[0]
                if hpid_to_observe.size > 0:
                    ras, decs = self._tesselate(hpid_to_observe)
                else:
                    ras = np.array([target_o_o.ra_rad_center])
                    decs = np.array([target_o_o.dec_rad_center])

                # Figure out an MJD start time for the object
                # if it is still rising and low.
                alt, az = _approx_ra_dec2_alt_az(
                    target_o_o.ra_rad_center,
                    target_o_o.dec_rad_center,
                    conditions.site.latitude_rad,
                    None,
                    conditions.mjd,
                    lmst=np.max(conditions.lmst),
                )
                HA = np.max(conditions.lmst) - target_o_o.ra_rad_center * 12.0 / np.pi

                if (HA < self.HA_max) & (HA > self.HA_min):
                    t_to_rise = (self.HA_max - HA) / 24.0
                    mjd0 = conditions.mjd + t_to_rise
                else:
                    mjd0 = conditions.mjd + 0.0

                obs_list = []
                for time, filternames, nv, exptime, index in zip(
                    self.times,
                    self.filters_at_times,
                    self.nvis,
                    self.exptimes,
                    np.arange(np.size(self.times)),
                ):
                    for i in range(nv):
                        # let's dither each pointing
                        if (i != 0) & (hpid_to_observe.size > 0):
                            ras, decs = self._tesselate(hpid_to_observe)

                        for filtername in filternames:
                            # Subsitute y for z if needed on first observation
                            if i == 0:
                                if (filtername == "z") & (filtername not in conditions.mounted_filters):
                                    filtername = "y"

                            if filtername == "u":
                                nexp = self.n_usnaps
                            else:
                                nexp = self.n_snaps

                            # If we are doing a short exposure
                            # need to be 1 snap for shutter limits
                            if exptime < 29.0:
                                nexp = 1

                            # check if we should break
                            # long exposures into multiple
                            if self.split_long:
                                if exptime > 119:
                                    nexp = int(np.round(exptime / 30.0))

                            obs = ScheduledObservationArray(ras.size)
                            obs["RA"] = ras
                            obs["dec"] = decs
                            obs["mjd"] = mjd0 + time
                            obs["flush_by_mjd"] = mjd0 + time + self.flushtime
                            obs["exptime"] = exptime
                            obs["nexp"] = nexp
                            obs["filter"] = filtername
                            obs["rotSkyPos"] = 0  # XXX--maybe throw a rotation detailer in here
                            obs["mjd_tol"] = self.mjd_tol
                            obs["dist_tol"] = self.dist_tol
                            obs["alt_min"] = self.alt_min
                            obs["alt_max"] = self.alt_max
                            obs["HA_max"] = self.HA_max
                            obs["HA_min"] = self.HA_min

                            obs["scheduler_note"] = self.survey_name + ", %i_t%i_i%i" % (
                                target_o_o.id,
                                time * 24,
                                index,
                            )
                            obs_list.append(obs)
                observations = np.concatenate(obs_list)
                self.set_script(observations)

    def calc_reward_function(self, conditions):
        """If there is an observation ready to go, execute it,
        otherwise, -inf"""
        # check if any new event has come in

        if conditions.targets_of_opportunity is not None:
            for target_o_o in conditions.targets_of_opportunity:
                if target_o_o.id > self.last_event_id:
                    self._new_event(target_o_o, conditions)
                    self.last_event_id = target_o_o.id

        observation = self.generate_observations_rough(conditions)

        if (observation is None) | (len(observation) == 0):
            self.reward = -np.inf
        else:
            self.reward = self.reward_val
        return self.reward

    def generate_observations(self, conditions):
        observations = self.generate_observations_rough(conditions)

        if len(observations) > 0:
            for detailer in self.detailers:
                observations = detailer(observations, conditions)

        return observations


def mean_longitude(longitude):
    """Compute a mean longitude, accounting for wrap around."""
    x = np.cos(longitude)
    y = np.sin(longitude)
    meanx = np.mean(x)
    meany = np.mean(y)
    angle = np.arctan2(meany, meanx)
    radius = np.sqrt(meanx**2 + meany**2)
    mid_longitude = angle % (2.0 * np.pi)
    if radius < 0.1:
        mid_longitude = np.pi
    return mid_longitude


def gen_too_surveys(
    nside=32,
    detailer_list=None,
    too_footprint=None,
    split_long=False,
    n_snaps=2,
):
    result = []

    ############
    # Generic GW followup
    ############

    # XXX---There's some extra stuff about do different things
    # if there's limited time before it sets. Let's see if this can
    # work first

    # XXX--instructions say do 4th night only 1/3 of the time.
    # Just leaving off for now

    times = [0, 24, 48]
    filters_at_times = ["ugrizy", "ugrizy", "ugrizy"]
    nvis = [3, 1, 1]
    exptimes = [120.0, 120.0, 120.0]
    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["GW_case_A"],
            survey_name="ToO, GW_case_A",
            split_long=split_long,
            flushtime=48.0,
            n_snaps=n_snaps,
        )
    )

    times = [0, 24, 48, 72]
    filters_at_times = ["gri", "ri", "ri", "ri"]
    nvis = [3, 1, 1, 1]
    exptimes = [120.0, 180.0, 180.0, 180.0]
    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["GW_case_B", "GW_case_C"],
            survey_name="ToO, GW_case_B_C",
            split_long=split_long,
            flushtime=48,
            n_snaps=n_snaps,
        )
    )

    times = [0, 24, 48, 72]
    filters_at_times = ["gr", "gr", "gr", "gr"]
    nvis = [1, 1, 1, 1]
    exptimes = [120.0, 120.0, 120.0, 120.0]
    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["GW_case_D", "GW_case_E"],
            survey_name="ToO, GW_case_D_E",
            split_long=split_long,
            flushtime=48,
            n_snaps=n_snaps,
        )
    )

    ############
    # Black hole-black hole GW merger
    ############

    # XXX--only considering bright objects now.

    times = np.array([0, 2, 7, 9, 39]) * 24
    filters_at_times = ["ugri"] * 5
    nvis = [1] * 5
    exptimes = [DEFAULT_EXP_TIME] * 5

    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["BBH_case_A", "BBH_case_B", "BBH_case_C"],
            survey_name="ToO, BBH",
            split_long=split_long,
            flushtime=48,
            n_snaps=n_snaps,
        )
    )

    ############
    # Lensed BNS
    ############

    times = [1.0, 1.0]
    filters_at_times = ["g", "r"]
    nvis = [1, 1]
    exptimes = [DEFAULT_EXP_TIME, 90.0]

    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["lensed_BNS_case_A"],
            survey_name="ToO, LensedBNS_A",
            split_long=split_long,
            flushtime=48.0,
            n_snaps=n_snaps,
        )
    )

    times = [1.0]
    filters_at_times = [
        "gr",
    ]
    nvis = [10]
    exptimes = [150]

    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["lensed_BNS_case_B"],
            survey_name="ToO, LensedBNS_B",
            split_long=split_long,
            flushtime=48.0,
            n_snaps=n_snaps,
        )
    )

    ############
    # Neutrino detector followup
    ############

    # XXX--need to update footprint to cut out galactic latitude

    times = [0, 0, 15 / 60.0, 0.5, 24, 24.5, 144]
    filters_at_times = ["g", "r", "z", "g", "r", "z", "grz"]
    exptimes = [
        120,
        DEFAULT_EXP_TIME,
        DEFAULT_EXP_TIME,
        120,
        DEFAULT_EXP_TIME,
        DEFAULT_EXP_TIME,
        DEFAULT_EXP_TIME,
    ]
    nvis = [1, 1, 1, 1, 1, 1, 1]

    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["neutrino"],
            survey_name="ToO, neutrino",
            split_long=split_long,
            flushtime=8.0,
            n_snaps=n_snaps,
        )
    )

    times = [0]
    filters_at_times = ["u"]
    exptimes = [DEFAULT_EXP_TIME]
    nvis = [1]

    # U-band with very long flush time.

    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["neutrino"],
            survey_name="ToO, neutrino_u",
            split_long=split_long,
            flushtime=1440,
            n_snaps=n_snaps,
        )
    )

    ############
    # Solar System
    ############
    # For the solar system objects, probably want a custom survey object,
    # but this should work for now. Want to add a detailer to add a dither
    # position.

    times = [0, 33 / 60.0, 66 / 60.0]
    filters_at_times = ["r"] * 3
    nvis = [1] * 3
    exptimes = [DEFAULT_EXP_TIME] * 3

    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["SSO_night"],
            survey_name="ToO, SSO_night",
            split_long=split_long,
            flushtime=2.0,
            n_snaps=n_snaps,
        )
    )

    times = [0, 10 / 60.0, 20 / 60.0]
    filters_at_times = ["z"] * 3
    nvis = [2] * 3
    exptimes = [15.0] * 3

    result.append(
        ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            filters_at_times=filters_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=detailer_list,
            too_types_to_follow=["SSO_twilight"],
            survey_name="ToO, SSO_twi",
            split_long=split_long,
            flushtime=2.0,
            n_snaps=n_snaps,
        )
    )

    return result



import unittest

import numpy as np

from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.schedulers import ComCamFilterSched, SimpleFilterSched
from rubin_scheduler.utils import survey_start_mjd


class TestFilterSchedulers(unittest.TestCase):

    def test_ComCamFilterSched(self):
        illum_bins = np.arange(0, 100 + 1, 50)
        filter_groups = (("g", "r", "i"), ("i", "z", "y"))
        filtersched = ComCamFilterSched(illum_bins=illum_bins, loaded_filter_groups=filter_groups)
        mjd = survey_start_mjd()
        conditions = Conditions(nside=8, mjd=mjd)
        conditions.moon_phase_sunset = 0
        load_filters = filtersched(conditions)
        self.assertTrue(load_filters == ["g", "r", "i"])
        conditions.moon_phase_sunset = 40
        load_filters = filtersched(conditions)
        self.assertTrue(load_filters == ["g", "r", "i"])
        conditions.moon_phase_sunset = 60
        load_filters = filtersched(conditions)
        self.assertTrue(load_filters == ["i", "z", "y"])
        conditions.moon_phase_sunset = 100
        load_filters = filtersched(conditions)
        self.assertTrue(load_filters == ["i", "z", "y"])

    def test_comcamfiltersched_except(self):
        illum_bins = np.arange(0, 100 + 1, 25)
        filter_groups = (("g", "r", "i"), ("i", "z", "y"))
        with self.assertRaises(ValueError):
            ComCamFilterSched(illum_bins=illum_bins, loaded_filter_groups=filter_groups)

    def test_SimpleFilterSched(self):
        filtersched = SimpleFilterSched(illum_limit=40)
        brightmoon_result = ["g", "r", "i", "z", "y"]
        newmoon_result = ["u", "g", "r", "i", "z"]
        mjd = survey_start_mjd()
        conditions = Conditions(nside=8, mjd=mjd)
        conditions.moon_phase_sunset = 0
        load_filters = filtersched(conditions)
        self.assertTrue(load_filters == newmoon_result)
        conditions.moon_phase_sunset = 50
        load_filters = filtersched(conditions)
        self.assertTrue(load_filters == brightmoon_result)


if __name__ == "__main__":
    unittest.main()



import importlib.util
import unittest

from rubin_scheduler.scheduler.schedulers.core_scheduler import CoreScheduler

if importlib.util.find_spec("lsst"):
    HAVE_TS = importlib.util.find_spec("lsst.ts")
else:
    HAVE_TS = False

if HAVE_TS:
    from rubin_scheduler.sim_archive import get_scheduler_instance_from_repo


class TestMakeSnapshot(unittest.TestCase):

    @unittest.skip("Skipping because test depends on external repo.")
    @unittest.skipIf(not HAVE_TS, "No lsst.ts")
    def test_get_scheduler_instance_photcal(self):
        scheduler = get_scheduler_instance_from_repo(
            config_repo="https://github.com/lsst-ts/ts_config_ocs.git",
            config_script="Scheduler/feature_scheduler/auxtel/fbs_config_image_photocal_survey.py",
            config_branch="main",
        )
        self.assertIsInstance(scheduler, CoreScheduler)



__all__ = ("sim_runner",)

import copy
import sqlite3
import sys
import time
import warnings

import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.schedulers import SimpleFilterSched
from rubin_scheduler.scheduler.utils import ObservationArray, SchemaConverter, run_info_table
from rubin_scheduler.utils import Site, _approx_altaz2pa, pseudo_parallactic_angle, rotation_converter


def sim_runner(
    observatory,
    scheduler,
    filter_scheduler=None,
    sim_start_mjd=None,
    sim_duration=3.0,
    filename=None,
    delete_past=True,
    n_visit_limit=None,
    step_none=15.0,
    verbose=True,
    extra_info=None,
    event_table=None,
    record_rewards=False,
    start_result_size=int(2e5),
    append_result_size=int(2.5e6),
    anomalous_overhead_func=None,
    telescope="rubin",
):
    """
    run a simulation

    Parameters
    ----------
    survey_length : float (3.)
        The length of the survey ot run (days)
    step_none : float (15)
        The amount of time to advance if the scheduler fails to
        return a target (minutes).
    extra_info : dict (None)
        If present, dict gets added onto the information from the
        observatory model.
    event_table : np.array (None)
        Any ToO events that were included in the simulation
    record_rewards : bool (False)
        Save computed rewards
    start_result_size : int
        Size of observations array to pre-allocate at the start of
        the run. Default 2e5.
    append_result_size : int
        Size of observations array to append if start_result_size is
        too small. Default 2.5e6.
    anomalous_overhead_func: `Callable` or None
        A function or callable object that takes the visit time and slew time
        (in seconds) as argument, and returns and additional offset (also
        in seconds) to be applied as addinional overhead between exposures.
        Defaults to None.
    telescope : `str`
        Name of telecope for camera rotation. Default "rubin".
    """

    if extra_info is None:
        extra_info = {}

    t0 = time.time()

    if filter_scheduler is None:
        filter_scheduler = SimpleFilterSched()

    if sim_start_mjd is None:
        mjd = observatory.mjd + 0
        sim_start_mjd = mjd + 0
    else:
        mjd = sim_start_mjd + 0
        observatory.mjd = mjd

    sim_end_mjd = sim_start_mjd + sim_duration
    observations = ObservationArray(n=start_result_size)
    mjd_track = mjd + 0
    step = 1.0 / 24.0
    step_none = step_none / 60.0 / 24.0  # to days
    mjd_run = sim_end_mjd - sim_start_mjd
    nskip = 0

    mjd_last_flush = -1

    last_obs_queue_fill_mjd_ns = None
    obs_rewards = {}
    reward_dfs = []

    rc = rotation_converter(telescope=telescope)

    # Make sure correct filters are mounted
    conditions = observatory.return_conditions()
    filters_needed = filter_scheduler(conditions)
    observatory.observatory.mount_filters(filters_needed)

    counter = 0
    while mjd < sim_end_mjd:
        if not scheduler._check_queue_mjd_only(observatory.mjd):
            scheduler.update_conditions(observatory.return_conditions())
        desired_obs = scheduler.request_observation(mjd=observatory.mjd)
        if record_rewards:
            if last_obs_queue_fill_mjd_ns != scheduler.queue_fill_mjd_ns:
                reward_dfs.append(scheduler.queue_reward_df)
                last_obs_queue_fill_mjd_ns = scheduler.queue_fill_mjd_ns

        if desired_obs is None:
            # No observation. Just step into the future and try again.
            warnings.warn("No observation. Step into the future and trying again.")
            observatory.mjd = observatory.mjd + step_none
            scheduler.update_conditions(observatory.return_conditions())
            nskip += 1
            continue
        completed_obs, new_night = observatory.observe(desired_obs)

        if completed_obs is not None:

            if anomalous_overhead_func is not None:
                observatory.mjd += (
                    anomalous_overhead_func(completed_obs["visittime"], completed_obs["slewtime"]) / 86400
                )

            scheduler.add_observation(completed_obs)
            observations[counter] = completed_obs[0]
            filter_scheduler.add_observation(completed_obs)
            counter += 1
            if counter == observations.size:
                add_observations = ObservationArray(n=append_result_size)
                observations = np.concatenate([observations, add_observations])

            if record_rewards:
                obs_rewards[completed_obs[0]["mjd"]] = last_obs_queue_fill_mjd_ns
        else:
            # An observation failed to execute, usually it was outside
            # the altitude limits.
            if observatory.mjd == mjd_last_flush:
                raise RuntimeError(
                    "Scheduler has failed to provide a valid observation multiple times "
                    f" at time ({observatory.mjd} from survey {scheduler.survey_index}."
                )
            # if this is a first offence, might just be that targets set.
            # Flush queue and try to get some new targets.
            scheduler.flush_queue()
            mjd_last_flush = copy.deepcopy(observatory.mjd)

        if new_night:
            # find out what filters we want mounted
            conditions = observatory.return_conditions()
            filters_needed = filter_scheduler(conditions)
            observatory.observatory.mount_filters(filters_needed)

        mjd = observatory.mjd + 0
        if verbose:
            if (mjd - mjd_track) > step:
                progress = np.max((mjd - sim_start_mjd) / mjd_run * 100)
                text = "\rprogress = %.2f%%" % progress
                sys.stdout.write(text)
                sys.stdout.flush()
                mjd_track = mjd + 0
        if n_visit_limit is not None:
            if counter == n_visit_limit:
                break
        # XXX--handy place to interupt and debug
        # if len(observations) > 25:
        #    import pdb ; pdb.set_trace()

    # trim off any observations that were pre-allocated but not used
    observations = observations[0:counter]

    # Compute alt,az,pa, rottelpos for observations
    # Only warn if it's a low-accuracy astropy conversion
    lsst = Site("LSST")

    # Using pseudo_parallactic_angle, see https://smtn-019.lsst.io/v/DM-44258/index.html
    pa, alt, az = pseudo_parallactic_angle(
        np.degrees(observations["RA"]),
        np.degrees(observations["dec"]),
        observations["mjd"],
        lon=lsst.longitude,
        lat=lsst.latitude,
        height=lsst.height,
    )
    observations["alt"] = np.radians(alt)
    observations["az"] = np.radians(az)
    observations["pseudo_pa"] = np.radians(pa)
    observations["rotTelPos"] = rc._rotskypos2rottelpos(observations["rotSkyPos"], observations["pseudo_pa"])

    # Also include traditional parallactic angle
    pa = _approx_altaz2pa(observations["alt"], observations["az"], lsst.latitude_rad)
    observations["pa"] = pa

    runtime = time.time() - t0
    print("Skipped %i observations" % nskip)
    print("Flushed %i observations from queue for being stale" % scheduler.flushed)
    print("Completed %i observations" % len(observations))
    print("ran in %i min = %.1f hours" % (runtime / 60.0, runtime / 3600.0))
    if len(observations) > 0:
        if filename is not None:
            print("Writing results to ", filename)
        if filename is not None:
            info = run_info_table(observatory, extra_info=extra_info)
            converter = SchemaConverter()
            converter.obs2opsim(observations, filename=filename, info=info, delete_past=delete_past)
        if event_table is not None:
            df = pd.DataFrame(event_table)
            con = sqlite3.connect(filename)
            df.to_sql("events", con)
            con.close()

        if record_rewards:
            reward_df = pd.concat(reward_dfs)
            obs_rewards_series = pd.Series(obs_rewards)
            obs_rewards_series.index.name = "mjd"
            obs_rewards_series.name = "queue_fill_mjd_ns"

            if filename is not None:
                with sqlite3.connect(filename) as con:
                    reward_df.to_sql("rewards", con)
                    obs_rewards_series.to_sql("obs_rewards", con)
    else:
        # Make sure there is something to return if there are no
        # observations
        reward_df = None
        obs_rewards_series = None

    if record_rewards:
        result = observatory, scheduler, observations, reward_df, obs_rewards_series
    else:
        result = observatory, scheduler, observations

    return result



import os
import unittest

import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler.utils import EuclidOverlapFootprint, SkyAreaGenerator, SkyAreaGeneratorGalplane

datadir = os.path.join(get_data_dir(), "scheduler")


class TestSkyArea(unittest.TestCase):
    def setUp(self):
        self.nside = 32

    @unittest.skipUnless(os.path.isdir(datadir), "Test data not available.")
    def test_skyareagenerator(self):
        # Just test that it sets up and returns maps
        s = SkyAreaGenerator(nside=self.nside)
        footprints, labels = s.return_maps()
        expected_labels = ["", "LMC_SMC", "bulge", "dusty_plane", "lowdust", "nes", "scp", "virgo"]
        self.assertEqual(set(np.unique(labels)), set(expected_labels))
        # Check that ratios in the low-dust wfd in r band are 1
        # This doesn't always have to be the case, but should be with defaults
        lowdust = np.where(labels == "lowdust")
        self.assertTrue(np.all(footprints["r"][lowdust] == 1))

    @unittest.skipUnless(os.path.isdir(datadir), "Test data not available.")
    def test_skyareagenerator_nside(self):
        # Just check two other likely common nsides
        for nside in (16, 64):
            s = SkyAreaGenerator(nside=nside)
            footprints, labels = s.return_maps()
            lowdust = np.where(labels == "lowdust")
            self.assertTrue(np.all(footprints["r"][lowdust] == 1))

    @unittest.skipUnless(os.path.isdir(datadir), "Test data not available.")
    def test_skyareageneratorgalplane(self):
        # Just test that it sets up and returns maps
        s = SkyAreaGeneratorGalplane(nside=self.nside)
        footprints, labels = s.return_maps()
        expected_labels = ["", "LMC_SMC", "bulgy", "dusty_plane", "lowdust", "nes", "scp", "virgo"]
        self.assertEqual(set(np.unique(labels)), set(expected_labels))
        # Check that ratios in the low-dust wfd in r band are 1
        # This doesn't always have to be the case, but should be with defaults
        lowdust = np.where(labels == "lowdust")
        self.assertTrue(np.all(footprints["r"][lowdust] == 1))

    @unittest.skipUnless(os.path.isdir(datadir), "Test data not available.")
    def test_euclidoverlapfootprint(self):
        # Just test that it sets up and returns maps
        s = EuclidOverlapFootprint(nside=self.nside)
        footprints, labels = s.return_maps()
        expected_labels = [
            "",
            "LMC_SMC",
            "bulgy",
            "dusty_plane",
            "lowdust",
            "nes",
            "scp",
            "virgo",
            "euclid_overlap",
        ]
        self.assertEqual(set(np.unique(labels)), set(expected_labels))
        # Check that ratios in the low-dust wfd in r band are 1
        # This doesn't always have to be the case, but should be with defaults
        lowdust = np.where(labels == "lowdust")
        self.assertTrue(np.all(footprints["r"][lowdust] == 1))


if __name__ == "__main__":
    unittest.main()



__all__ = ["gen_roman_on_season", "gen_roman_off_season"]

import warnings

import rubin_scheduler.scheduler.basis_functions as bf
import rubin_scheduler.scheduler.detailers as detailers
from rubin_scheduler.scheduler.surveys import DeepDrillingSurvey


def roman_info():
    """Manually enter some Roman RGES info."""
    # From the TVS Slack channel:
    # spring 2027, fall 2027, spring 2028, then fall 2030,
    # spring 2031, fall 2031

    result = {}
    result["RA"] = 268.708
    result["dec"] = -28.975

    # Guessing these from the notebook in same dir.
    observing_season_mid_mjds = [61947.3, 62318.8, 62670.3, 63067.2, 63381.4, 63773.2]

    result["seasons_on"] = [[val - 32, val + 32] for val in observing_season_mid_mjds]

    result["seasons_off"] = []
    for i in range(len(result["seasons_on"]) - 1):
        result["seasons_off"].append([result["seasons_on"][i][1] + 1, result["seasons_on"][i + 1][0] - 1])

    return result


def gen_roman_on_season(
    nside=32,
    camera_ddf_rot_limit=75.0,
    exptime=30.0,
    nexp=2,
):
    """Generate a survey object for observing the Roman field(s)
    in an on season"""

    warnings.warn("Generating Roman survey place holder. Should probably not be in production.")

    field_info = roman_info()

    RA = field_info["RA"]
    dec = field_info["dec"]

    survey_name = "DD: RGES_onseason"

    # Add some feasability basis functions. Maybe just give it a set
    # of nights where it can execute for now.
    basis_functions = []
    # These are crude hard limits. Nominally we would try to
    # pre-schedule these when they would be at the best airamss
    # in the night.
    basis_functions.append(bf.HourAngleLimitBasisFunction(RA=RA, ha_limits=[[20, 24], [0, 4]]))
    basis_functions.append(bf.NotTwilightBasisFunction())
    # Force it to delay about a day
    basis_functions.append(bf.ForceDelayBasisFunction(days_delay=0.8, survey_name=survey_name))
    # Force it to be in a given observing season
    basis_functions.append(bf.InTimeWindowBasisFunction(mjd_windows=field_info["seasons_on"]))
    basis_functions.append(bf.MoonDistPointRangeBasisFunction(RA, dec))
    basis_functions.append(bf.AirmassPointRangeBasisFunction(RA, dec, nside=nside))

    # Add a dither detailer, so it dithers between each set
    # of exposures I guess?
    details = []
    details.append(detailers.DitherDetailer(max_dither=0.5, seed=42, per_night=True))
    details.append(detailers.CameraRotDetailer(min_rot=-camera_ddf_rot_limit, max_rot=camera_ddf_rot_limit))

    survey = DeepDrillingSurvey(
        basis_functions,
        RA=RA,
        dec=dec,
        sequence="giriz",
        nvis=[1, 1, 1, 1, 1],
        exptime=exptime,
        nexp=nexp,
        survey_name=survey_name,
        detailers=details,
    )
    return survey


def gen_roman_off_season(
    nside=32,
    camera_ddf_rot_limit=75.0,
    exptime=30.0,
    nexp=2,
):
    """Generate a ddf-like survey object to observe the roman
    field every ~3 days in the off-season"""

    warnings.warn("Generating Roman survey place holder. Should probably not be in production.")

    field_info = roman_info()
    RA = field_info["RA"]
    dec = field_info["dec"]

    survey_name = "DD: RGES_offseason"

    # Add some feasability basis functions. Maybe just give it a
    # set of nights where it can execute for now.
    basis_functions = []
    # These are crude hard limits. Nominally we would try
    # to pre-schedule these when they would be at the best
    # airamss in the night.
    basis_functions.append(bf.HourAngleLimitBasisFunction(RA=RA, ha_limits=[[20, 24], [0, 4]]))
    basis_functions.append(bf.NotTwilightBasisFunction())
    # Force it to not go every day
    basis_functions.append(bf.ForceDelayBasisFunction(days_delay=3.0, survey_name=survey_name))
    # Force it to be in a given observing season
    basis_functions.append(bf.InTimeWindowBasisFunction(mjd_windows=field_info["seasons_off"]))
    basis_functions.append(bf.MoonDistPointRangeBasisFunction(RA, dec))
    basis_functions.append(bf.AirmassPointRangeBasisFunction(RA, dec, nside=nside))

    # Add a dither detailer, so it dithers between each
    # set of exposures I guess?
    details = []
    details.append(detailers.DitherDetailer(max_dither=0.5, seed=42, per_night=True))
    details.append(detailers.CameraRotDetailer(min_rot=-camera_ddf_rot_limit, max_rot=camera_ddf_rot_limit))

    survey = DeepDrillingSurvey(
        basis_functions,
        RA=RA,
        dec=dec,
        sequence="griz",
        nvis=[1, 1, 1, 1],
        exptime=exptime,
        nexp=nexp,
        survey_name=survey_name,
        detailers=details,
    )
    return survey



__all__ = (
    "FilterLoadedBasisFunction",
    "OnceInNightBasisFunction",
    "SunAltHighLimitBasisFunction",
    "TimeToTwilightBasisFunction",
    "NotTwilightBasisFunction",
    "AfterEveningTwiBasisFunction",
    "ForceDelayBasisFunction",
    "HourAngleLimitBasisFunction",
    "MoonDownBasisFunction",
    "FractionOfObsBasisFunction",
    "CloudedOutBasisFunction",
    "RisingMoreBasisFunction",
    "SoftDelayBasisFunction",
    "LookAheadDdfBasisFunction",
    "SunAltLimitBasisFunction",
    "TimeInTwilightBasisFunction",
    "NightModuloBasisFunction",
    "EndOfEveningBasisFunction",
    "TimeToScheduledBasisFunction",
    "CloseToTwilightBasisFunction",
    "MoonDistPointRangeBasisFunction",
    "AirmassPointRangeBasisFunction",
    "InTimeWindowBasisFunction",
)

import warnings

import numpy as np

from rubin_scheduler.scheduler import features
from rubin_scheduler.scheduler.basis_functions import BaseBasisFunction
from rubin_scheduler.scheduler.utils import IntRounded
from rubin_scheduler.utils import _angular_separation, ra_dec2_hpid


def send_unused_deprecation_warning(name):
    message = (
        f"The feasibility basis function {name} is not in use, "
        "may be broken, and will be deprecated shortly. "
        "Please contact the rubin_scheduler maintainers if "
        "this is in use elsewhere."
    )
    warnings.warn(message, FutureWarning)


class FilterLoadedBasisFunction(BaseBasisFunction):
    """Check that the filter(s) needed are loaded.

    Are the filters in `filternames` loaded and available?

    Parameters
    ----------
    filternames : `str` or `list` [ `str` ]
        The filternames that need to be mounted to execute.
    """

    def __init__(self, filternames="r"):
        super(FilterLoadedBasisFunction, self).__init__()
        if not isinstance(filternames, list):
            filternames = [filternames]
        self.filternames = filternames

    def check_feasibility(self, conditions):
        for filtername in self.filternames:
            result = filtername in conditions.mounted_filters
            if result is False:
                return result
        return result


class InTimeWindowBasisFunction(BaseBasisFunction):
    """Only let a survey go if it is in a defined season

    Parameters
    ----------
    mjd_windows : `list`
        List of mjd pairs. feasability will only pass if the
        current MJD falls in a viable range, e.g.
        seaons=[[1,2], [10,13]], means observations can happen
        on mjd=1,2,10,11,12,13.
    """

    def __init__(self, mjd_windows=[]):
        super().__init__()
        self.mjd_windows = mjd_windows

    def check_feasibility(self, conditions):
        result = False
        for mjd_windows in self.mjd_windows:
            if np.min(mjd_windows) <= conditions.mjd <= np.max(mjd_windows):
                result = True
        return result


class AirmassPointRangeBasisFunction(BaseBasisFunction):
    """Set an airmass limit for a single point
    Maps the point to the nearest HEALpixel and uses
    the Conditions airmass map, so precision of the
    airmass limit that is applied is limited by the
    HEALpix resolution.

    Parameters
    ----------
    ra : `float`
        The RA of the point (degrees)
    dec : `float`
        The Dec of the point (degrees)
    airmass_range : `list`
        The valid airmass range, default [1.05, 2.7].
    """

    def __init__(self, ra, dec, airmass_range=[1.05, 2.7], nside=32):
        super().__init__()
        self.hpid = ra_dec2_hpid(nside, ra, dec)
        self.airmass_range = airmass_range

    def check_feasibility(self, conditions):
        result = False
        airmass = conditions.airmass[self.hpid]
        if (np.min(self.airmass_range) <= airmass) & (airmass <= np.max(self.airmass_range)):
            result = True
        return result


class MoonDistPointRangeBasisFunction(BaseBasisFunction):
    """Set a moon distance limit for a single point.

    Parameters
    ----------
    ra : `float`
        The RA of the point (degrees)
    dec : `float`
        The Dec of the point (degrees)
    moon_limit : `float`
        The angular distance to demand from the moon (degrees).
        Default 15.
    """

    def __init__(self, ra, dec, moon_limit=15.0):
        super().__init__()
        self.ra = np.radians(ra)
        self.dec = np.radians(dec)
        self.moon_limit = np.radians(moon_limit)

    def check_feasibility(self, conditions):
        result = False
        moon_dist = _angular_separation(self.ra, self.dec, conditions.moon_ra, conditions.moon_dec)
        if moon_dist > self.moon_limit:
            result = True
        return result


class CloseToTwilightBasisFunction(BaseBasisFunction):
    """Only execute if the sun is higher than `sun_alt_limit`,
    the current time is within `time_to_12deg` of -12 degree twilight,
    and there is at least `time_remaining` time left before -12 degree
    twilight.

    Is the current time and sun altitude close to twilight, but not too close?

    Parameters
    ----------
    max_sun_alt_limit : `float`
        The sun altitude limit (degrees). Sun must be higher than this
        at sunset to execute
    max_time_to_12deg : `float`
        How much time must be remaining before 12 degree twilight in
        the morning (minutes)
    min_time_remaining : `float`
        Minimum about of time that must be available before trying to
        execute (minutes)

    Notes
    -----
    This is primarily useful for surveys which must execute within close
    limits of -12 degree twilight.
    """

    def __init__(self, max_sun_alt_limit=-14.8, max_time_to_12deg=21.0, min_time_remaining=15.0):
        super().__init__()
        if max_time_to_12deg < min_time_remaining:
            raise ValueError(
                "time_to_12deg value of %f is less than time_remaining value of %f."
                % (max_time_to_12deg, min_time_remaining)
            )
        self.sun_alt_limit = np.radians(max_sun_alt_limit)
        self.time_to_12deg = max_time_to_12deg / 60.0 / 24.0
        self.min_time_remaining = min_time_remaining / 60.0 / 24.0

    def check_feasibility(self, conditions):
        result = False
        # If the sun is high, it's ok to execute
        if conditions.sun_alt > self.sun_alt_limit:
            result = True
        time_left = conditions.sun_n12_rising - conditions.mjd
        if time_left < self.time_to_12deg:
            result = True
        if time_left < self.min_time_remaining:
            result = False
        return result


class OnceInNightBasisFunction(BaseBasisFunction):
    """Stop observing if something has been executed already in the night

    Parameters
    ----------
    notes : `list` [ `str` ]
        A list of str to check if any observations with a matching note exist.
    """

    def __init__(self, notes=[]):
        super(OnceInNightBasisFunction, self).__init__()
        self.survey_features["note_in_night"] = features.NoteInNight(notes=notes)

    def check_feasibility(self, conditions):
        result = True
        if self.survey_features["note_in_night"].feature > 0:
            result = False
        return result


class SunAltHighLimitBasisFunction(BaseBasisFunction):
    """Don't observe unless the sun is above some limit

    Parameters
    ----------
    alt_limit : float (-15)
        Sun must be above alt_limit (degrees). Default -15.
    """

    def __init__(self, alt_limit=-15.0):
        super(SunAltHighLimitBasisFunction, self).__init__()
        self.alt_limit = np.radians(alt_limit)

    def check_feasibility(self, conditions):
        result = True
        if conditions.sun_alt < self.alt_limit:
            result = False
        return result


class NightModuloBasisFunction(BaseBasisFunction):
    """Only return true on certain nights"""

    def __init__(self, pattern=None):
        super(NightModuloBasisFunction, self).__init__()
        if pattern is None:
            pattern = [True, False]
        self.pattern = pattern
        self.mod_val = len(self.pattern)

    def check_feasibility(self, conditions):
        indx = int(conditions.night % self.mod_val)
        result = self.pattern[indx]
        return result


class TimeInTwilightBasisFunction(BaseBasisFunction):
    """Make sure there is some time left in twilight.

    Parameters
    ----------
    time_needed : float (5)
        The time needed remaining in twilight (minutes)
    """

    def __init__(self, time_needed=5.0):
        super(TimeInTwilightBasisFunction, self).__init__()
        self.time_needed = time_needed / 60.0 / 24.0  # To days

    def check_feasibility(self, conditions):
        result = False
        time1 = conditions.sun_n18_setting - conditions.mjd
        time2 = conditions.sun_n12_rising - conditions.mjd

        if time1 > self.time_needed:
            result = True
        else:
            if conditions.sunAlt > np.radians(-18.0):
                if time2 > self.time_needed:
                    result = True
        return result


class AfterEveningTwiBasisFunction(BaseBasisFunction):
    """Only execute right after evening twilight"""

    def __init__(self, time_after=30.0, alt_limit=18):
        super(AfterEveningTwiBasisFunction, self).__init__()
        self.time_after = IntRounded(time_after / 60.0 / 24.0)
        self.alt_limit = str(alt_limit)

    def check_feasibility(self, conditions):
        available_time = conditions.mjd - getattr(conditions, "sun_n" + self.alt_limit + "_setting")
        result = IntRounded(available_time) < self.time_after
        return result


class EndOfEveningBasisFunction(BaseBasisFunction):
    """Only let observations happen in a limited time before twilight"""

    def __init__(self, time_remaining=30.0, alt_limit=18):
        super(EndOfEveningBasisFunction, self).__init__()
        self.time_remaining = IntRounded(time_remaining / 60.0 / 24.0)
        self.alt_limit = str(alt_limit)

    def check_feasibility(self, conditions):
        available_time = getattr(conditions, "sun_n" + self.alt_limit + "_rising") - conditions.mjd
        result = IntRounded(available_time) < self.time_remaining
        return result


class TimeToTwilightBasisFunction(BaseBasisFunction):
    """Make sure there is enough time before twilight. Useful
    if you want to check before starting a long sequence of observations.

    Parameters
    ----------
    time_needed : float (30.)
        The time needed to run a survey (mintues).
    alt_limit : int (18)
        The sun altitude limit to use. Must be 12 or 18
    """

    def __init__(self, time_needed=30.0, alt_limit=18):
        super(TimeToTwilightBasisFunction, self).__init__()
        self.time_needed = time_needed / 60.0 / 24.0  # To days
        self.alt_limit = str(alt_limit)

    def check_feasibility(self, conditions):
        available_time = getattr(conditions, "sun_n" + self.alt_limit + "_rising") - conditions.mjd
        result = available_time > self.time_needed
        return result


class TimeToScheduledBasisFunction(BaseBasisFunction):
    """Make sure there is enough time before next scheduled observation.
    Useful if you want to check before starting a long sequence of
    observations.

    Parameters
    ----------
    time_needed : float (30.)
        The time needed to run a survey (mintues).
    """

    def __init__(self, time_needed=30.0):
        super(TimeToScheduledBasisFunction, self).__init__()
        self.time_needed = time_needed / 60.0 / 24.0  # To days

    def check_feasibility(self, conditions):
        if len(conditions.scheduled_observations) == 0:
            return True

        available_time = np.min(conditions.scheduled_observations) - conditions.mjd
        result = available_time > self.time_needed
        return result


class NotTwilightBasisFunction(BaseBasisFunction):
    def __init__(self, sun_alt_limit=-18):
        """
        # Should be -18 or -12
        """
        self.sun_alt_limit = str(int(sun_alt_limit)).replace("-", "n")
        super(NotTwilightBasisFunction, self).__init__()

    def check_feasibility(self, conditions):
        result = True
        if conditions.mjd < getattr(conditions, "sun_" + self.sun_alt_limit + "_setting"):
            result = False
        if conditions.mjd > getattr(conditions, "sun_" + self.sun_alt_limit + "_rising"):
            result = False
        return result


class ForceDelayBasisFunction(BaseBasisFunction):
    """Keep a survey from executing too rapidly.

    Parameters
    ----------
    days_delay : `float`, optional
        The number of days to force a gap on.
    scheduler_note : `str` or None, optional
        The value of the scheduler_note to count.
        Default None will not consider scheduler_note.
    survey_name : `str` or None, optional
        Backwards compatible version of scheduler_note.
    """

    def __init__(self, days_delay=2.0, scheduler_note=None, survey_name=None):
        super(ForceDelayBasisFunction, self).__init__()
        self.days_delay = days_delay
        if scheduler_note is None and survey_name is not None:
            self.scheduler_note = survey_name
        else:
            self.scheduler_note = scheduler_note
        self.survey_features["last_obs_self"] = features.LastObservation(scheduler_note=self.scheduler_note)

    def check_feasibility(self, conditions):
        result = True
        if conditions.mjd - self.survey_features["last_obs_self"].feature["mjd"] < self.days_delay:
            result = False
        return result


class SoftDelayBasisFunction(BaseBasisFunction):
    """Like Force_delay, but go ahead and let things catch up if they fall
    far behind.

    Parameters
    ----------
    fractions : `list` [`float`]
    delays : `list` [`float`]
    scheduler_note : `str` or None, optional
        The scheduler_note to identify observations from a given survey or
        survey mode.
    survey_name : `str` or None, optional
        Deprecated version of scheduler_note. Overriden by scheduler_note
        if not None.
    """

    def __init__(
        self, fractions=[0.000, 0.009, 0.017], delays=[0.0, 0.5, 1.5], scheduler_note=None, survey_name=None
    ):
        if len(fractions) != len(delays):
            raise ValueError("fractions and delays must be same length")
        super(SoftDelayBasisFunction, self).__init__()
        self.delays = delays
        if scheduler_note is None and survey_name is not None:
            self.scheduler_note = survey_name
        else:
            self.scheduler_note = scheduler_note
        self.survey_features["last_obs_self"] = features.LastObservation(scheduler_note=self.scheduler_note)
        self.fractions = fractions
        self.survey_features["N_total"] = features.NObsCount(note=None)
        self.survey_features["N_note"] = features.NObsCount(note=self.scheduler_note)

    def check_feasibility(self, conditions):
        result = True
        current_ratio = self.survey_features["N_note"].feature / self.survey_features["N_total"].feature
        indx = np.searchsorted(self.fractions, current_ratio)
        if indx == len(self.fractions):
            indx -= 1
        delay = self.delays[indx]
        if conditions.mjd - self.survey_features["last_obs_self"].feature["mjd"] < delay:
            result = False
        return result


class HourAngleLimitBasisFunction(BaseBasisFunction):
    """Only execute a survey in limited hour angle ranges. Useful for
    limiting Deep Drilling Fields.

    Parameters
    ----------
    RA : float (0.)
        RA of the target (degrees).
    ha_limits : list of lists
        limits for what hour angles are acceptable (hours). e.g.,
        to give 4 hour window around RA=0, ha_limits=[[22,24], [0,2]]
    """

    def __init__(self, RA=0.0, ha_limits=None):
        super(HourAngleLimitBasisFunction, self).__init__()
        self.ra_hours = RA / 360.0 * 24.0
        self.ha_limits = np.array(ha_limits)

    def check_feasibility(self, conditions):
        target_ha = (conditions.lmst - self.ra_hours) % 24
        # Are we in any of the possible windows
        result = False
        for limit in self.ha_limits:
            lres = limit[0] <= target_ha < limit[1]
            result = result or lres

        return result


class MoonDownBasisFunction(BaseBasisFunction):
    """Demand the moon is down"""

    def check_feasibility(self, conditions):
        result = True
        if conditions.moon_alt > 0:
            result = False
        return result


class FractionOfObsBasisFunction(BaseBasisFunction):
    """Limit the fraction of all observations that can be labelled a certain
    scheduler note.

    Useful for keeping DDFs from exceeding a given fraction
    of the total survey.

    Parameters
    ----------
    frac_total : `float`
        The fraction of total observations that can be of this survey
    scheduler_note : `str` or None, optional
        The scheduler_note to identify observations from a given survey or
        survey mode.
    survey_name : `str` or None, optional
        Deprecated version of scheduler_note. Overriden by scheduler_note
        if scheduler_note not None.
    """

    def __init__(self, frac_total, scheduler_note=None, survey_name=None):
        super(FractionOfObsBasisFunction, self).__init__()
        if scheduler_note is None and survey_name is not None:
            self.scheduler_note = survey_name
        else:
            self.scheduler_note = scheduler_note
        self.frac_total = frac_total
        self.survey_features["N_total"] = features.NObsCount(note=None)
        self.survey_features["N_note"] = features.NObsCount(note=self.scheduler_note)

    def check_feasibility(self, conditions):
        # If nothing has been observed, fine to go
        result = True
        if self.survey_features["N_total"].feature > 0:
            ratio = self.survey_features["N_note"].feature / self.survey_features["N_total"].feature
            if ratio > self.frac_total:
                result = False
        return result


class LookAheadDdfBasisFunction(BaseBasisFunction):
    """Look into the future to decide if it's a good time to observe or block.

    Parameters
    ----------
    frac_total : float
        The fraction of total observations that can be of this survey
    aggressive_fraction : float
        If the fraction of observations drops below ths value, be more
        aggressive in scheduling.  e.g., do not wait for conditions to
        improve, execute as soon as possible.
    time_needed : float (30.)
        Estimate of the amount of time needed to execute DDF sequence
        (minutes).
    RA : float (0.)
        The RA of the DDF
    ha_limits : list of lists (None)
        limits for what hour angles are acceptable (hours). e.g.,
        to give 4 hour window around HA=0, ha_limits=[[22,24], [0,2]]
    scheduler_note : `str` or None, optional
        The scheduler_note to identify observations from a given survey or
        survey mode.
    survey_name : `str` or None, optional
        Deprecated version of scheduler_note. Overriden by scheduler_note
        if scheduler_note not None.
    time_jump : float (44.)
        The amount of time to assume will jump ahead if another survey
        executes (minutes)
    sun_alt_limit : float (-18.)
        The limit to assume twilight starts (degrees)
    """

    def __init__(
        self,
        frac_total,
        aggressive_fraction,
        time_needed=30.0,
        RA=0.0,
        ha_limits=None,
        scheduler_note=None,
        survey_name=None,
        time_jump=44.0,
        sun_alt_limit=-18.0,
    ):
        super(LookAheadDdfBasisFunction, self).__init__()
        if aggressive_fraction > frac_total:
            raise ValueError("aggressive_fraction should be less than frac_total")
        if scheduler_note is None and survey_name is not None:
            self.scheduler_note = survey_name
        else:
            self.scheduler_note = scheduler_note
        self.frac_total = frac_total
        self.ra_hours = RA / 360.0 * 24.0
        self.ha_limits = np.array(ha_limits)
        self.sun_alt_limit = str(int(sun_alt_limit)).replace("-", "n")
        self.time_jump = time_jump / 60.0 / 24.0  # To days
        self.time_needed = time_needed / 60.0 / 24.0  # To days
        self.aggressive_fraction = aggressive_fraction
        self.survey_features["N_total"] = features.NObsCount(note=None)
        self.survey_features["N_note"] = features.NObsCount(note=self.scheduler_note)

    def check_feasibility(self, conditions):
        result = True
        target_ha = (conditions.lmst - self.ra_hours) % 24
        ratio = self.survey_features["N_note"].feature / self.survey_features["N_total"].feature
        available_time = getattr(conditions, "sun_" + self.sun_alt_limit + "_rising") - conditions.mjd
        # If it's more that self.time_jump to hour angle zero
        # See if there will be enough time to twilight in the future
        if (IntRounded(target_ha) > IntRounded(12)) & (
            IntRounded(target_ha) < IntRounded(24.0 - self.time_jump)
        ):
            if IntRounded(available_time) > IntRounded(self.time_needed + self.time_jump):
                result = False
                # If we paused for better conditions, but the moon will
                # rise, turn things back on.
                if IntRounded(conditions.moon_alt) < IntRounded(0):
                    if IntRounded(conditions.moonrise) > IntRounded(conditions.mjd):
                        if IntRounded(conditions.moonrise - conditions.mjd) > IntRounded(self.time_jump):
                            result = True
        # If the moon is up and will set soon, pause
        if IntRounded(conditions.moon_alt) > IntRounded(0):
            time_after_moonset = (
                getattr(conditions, "sun_" + self.sun_alt_limit + "_rising") - conditions.moonset
            )
            if IntRounded(conditions.moonset) > IntRounded(self.time_jump):
                if IntRounded(time_after_moonset) > IntRounded(self.time_needed):
                    result = False

        # If the survey has fallen far behind, be agressive and observe
        # anytime it's up.
        if IntRounded(ratio) < IntRounded(self.aggressive_fraction):
            result = True
        return result


class CloudedOutBasisFunction(BaseBasisFunction):
    def __init__(self, cloud_limit=0.7):
        super(CloudedOutBasisFunction, self).__init__()
        self.cloud_limit = cloud_limit

    def check_feasibility(self, conditions):
        result = True
        if conditions.bulk_cloud > self.cloud_limit:
            result = False
        return result


class RisingMoreBasisFunction(BaseBasisFunction):
    """Say a spot is not available if it will rise substantially before
    twilight.

    Parameters
    ----------
    RA : float
        The RA of the point in the sky (degrees)
    pad : float
        When to start observations if there's plenty of time before
        twilight (minutes)
    """

    def __init__(self, RA, pad=30.0):
        super(RisingMoreBasisFunction, self).__init__()
        self.ra_hours = RA * 24 / 360.0
        self.pad = pad / 60.0  # To hours

    def check_feasibility(self, conditions):
        result = True
        # Calculating hour angle instead of using conditions.HA
        # because want -12 to +12 hours instead of 0 - 2pi values.
        hour_angle = conditions.lmst - self.ra_hours
        # If it's rising, and twilight is well beyond when it crosses
        # the meridian
        time_to_twi = (conditions.sun_n18_rising - conditions.mjd) * 24.0
        if (hour_angle < -self.pad) & (np.abs(hour_angle) < (time_to_twi - self.pad)):
            result = False
        return result


class SunAltLimitBasisFunction(BaseBasisFunction):
    """Only observe if the sun is below a given altitude limit.

    Parameters
    ----------
    alt_limit : `float`
        The maximum altitude for the sun.
    """

    def __init__(self, alt_limit=-12.1):
        super(SunAltLimitBasisFunction, self).__init__()
        self.alt_limit = np.radians(alt_limit)

    def check_feasibility(self, conditions):
        result = True
        if conditions.sun_alt > self.alt_limit:
            result = False
        return result



__all__ = (
    "BaseDetailer",
    "ZeroRotDetailer",
    "Comcam90rotDetailer",
    "Rottep2RotspDesiredDetailer",
    "CloseAltDetailer",
    "TakeAsPairsDetailer",
    "TwilightTripleDetailer",
    "FlushForSchedDetailer",
    "FilterNexp",
    "FixedSkyAngleDetailer",
    "ParallacticRotationDetailer",
    "FlushByDetailer",
    "RandomFilterDetailer",
    "TrackingInfoDetailer",
)

import copy

import numpy as np

from rubin_scheduler.scheduler.utils import IntRounded
from rubin_scheduler.utils import (
    _angular_separation,
    _approx_altaz2pa,
    _approx_ra_dec2_alt_az,
    pseudo_parallactic_angle,
    rotation_converter,
)


class BaseDetailer:
    """
    A Detailer is an object that takes a list of proposed observations and
    adds "details" to them. The primary purpose is that the Markov Decision
    Process does an excelent job selecting RA,Dec,filter combinations, but we
    may want to add additional logic such as what to set the camera rotation
    angle to, or what to use for an exposure time. We could also modify the
    order of the proposed observations. For Deep Drilling Fields, a detailer
    could be useful for computing dither positions and modifying the exact
    RA,Dec positions.
    """

    def __init__(self, nside=32):
        """"""
        # Dict to hold all the features we want to track
        self.survey_features = {}
        self.nside = nside

    def add_observations_array(self, observations_array, observations_hpid):
        """Like add_observation, but for loading a whole array of
        observations at a time"""

        for feature in self.survey_features:
            self.survey_features[feature].add_observations_array(observations_array, observations_hpid)

    def add_observation(self, observation, indx=None):
        """
        Parameters
        ----------
        observation : `np.array`
            An array with information about the input observation
        indx : `np.array`
            The indices of the healpix map that the observation overlaps
            with
        """
        for feature in self.survey_features:
            self.survey_features[feature].add_observation(observation, indx=indx)

    def __call__(self, observation_list, conditions):
        """
        Parameters
        ----------
        observation_list : `list` of observations
            The observations to detail.
        conditions : `rubin_scheduler.scheduler.conditions` object

        Returns
        -------
        List of observations.
        """

        return observation_list


class TrackingInfoDetailer(BaseDetailer):
    """Fill in lots of the different tracking strings for an observation."""

    def __init__(self, target_name=None, science_program=None, observation_reason=None):
        self.survey_features = {}
        self.target_name = target_name
        self.science_program = science_program
        self.observation_reason = observation_reason

    def __call__(self, observation_list, conditions):
        for obs in observation_list:
            if self.science_program is not None:
                obs["science_program"] = self.science_program
            if self.target_name is not None:
                obs["target_name"] = self.target_name
            if self.observation_reason is not None:
                obs["observation_reason"] = self.observation_reason

        return observation_list


class FlushByDetailer(BaseDetailer):
    """Set the MJD an observation should be flushed from the scheduler
    queue if not yet completed.

    Parameters
    ----------
    flush_time : float
        The time to flush after the current MJD. Default 60 minutes
    """

    def __init__(self, flush_time=60, nside=32):
        self.survey_features = {}
        self.nside = nside
        self.flush_time = flush_time / 60 / 24.0

    def __call__(self, observation_list, conditions):
        for obs in observation_list:
            obs["flush_by_mjd"] = conditions.mjd + self.flush_time
        return observation_list


class RandomFilterDetailer(BaseDetailer):
    """Pick a random filter for the observations

    Parameters
    ----------
    filters : `str`
        The filters to randomize. Default 'riz'
    nights_to_prep : `int`
        The number of nights to generate random filters for.
        Default 10000.
    seed : number
        Seed for RNG. Defaut 42
    fallback_order : `str`
        If the desired filter is not mounted, goes through
        `fallback_order` and uses the first filter that is
        available
    """

    def __init__(self, filters="riz", nights_to_prep=10000, seed=42, fallback_order="rizgyu"):
        self.survey_features = []
        self.fallback_order = fallback_order

        rng = np.random.default_rng(seed)
        self.night2filter_int = rng.integers(low=0, high=len(filters), size=nights_to_prep)

        self.filter_dict = {}
        for i, filtername in enumerate(filters):
            self.filter_dict[i] = filtername

    def __call__(self, observation_list, conditions):

        filter_to_use = self.filter_dict[self.night2filter_int[conditions.night]]
        # Filter not available
        if filter_to_use not in conditions.mounted_filters:
            is_mounted = [filtername in conditions.mounted_filters for filtername in self.fallback_order]
            indx = np.min(np.where(is_mounted))
            filter_to_use = self.fallback_order[indx]

        for obs in observation_list:
            obs["filter"] = filter_to_use
        return observation_list


class ParallacticRotationDetailer(BaseDetailer):
    """Set the rotator to near the parallactic angle"""

    def __init__(self, telescope="rubin"):
        self.rc = rotation_converter(telescope=telescope)
        self.survey_features = {}

    def __call__(self, observation_list, conditions, limits=[-270, 270]):
        limits = np.radians(limits)
        for obs in observation_list:
            alt, az = _approx_ra_dec2_alt_az(
                obs["RA"],
                obs["dec"],
                conditions.site.latitude_rad,
                conditions.site.longitude_rad,
                conditions.mjd,
            )
            obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)
            obs["rotSkyPos_desired"] = obs_pa

            resulting_rot_tel_pos = self.rc._rotskypos2rottelpos(obs["rotSkyPos_desired"], obs_pa)

            if resulting_rot_tel_pos > np.max(limits):
                resulting_rot_tel_pos -= 2 * np.pi
            if resulting_rot_tel_pos < np.min(limits):
                resulting_rot_tel_pos += 2 * np.pi

            # If those corrections still leave us bad, just pull it back 180.
            if resulting_rot_tel_pos > np.max(limits):
                resulting_rot_tel_pos -= np.pi

            # The rotTelPos overides everything else.
            obs["rotTelPos"] = resulting_rot_tel_pos
            # if the rotSkyPos_desired isn't possible, fall back to this.
            obs["rotTelPos_backup"] = 0

        return observation_list


class Rottep2RotspDesiredDetailer(BaseDetailer):
    """Convert all the rotTelPos values to rotSkyPos_desired"""

    def __init__(self, telescope="rubin"):
        self.rc = rotation_converter(telescope=telescope)
        self.survey_features = {}

    def __call__(self, observation_list, conditions):
        obs_array = np.concatenate(observation_list)

        alt, az = _approx_ra_dec2_alt_az(
            obs_array["RA"],
            obs_array["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)

        rot_sky_pos_desired = self.rc._rotskypos2rottelpos(obs_array["rotTelPos"], obs_pa)

        for obs, rotsp_d in zip(observation_list, rot_sky_pos_desired):
            obs["rotTelPos_backup"] = obs["rotTelPos"] + 0
            obs["rotTelPos"] = np.nan
            obs["rotSkyPos"] = np.nan
            obs["rotSkyPos_desired"] = rotsp_d

        return observation_list


class ZeroRotDetailer(BaseDetailer):
    """
    Detailer to set the camera rotation to be apporximately zero in
    rotTelPos.

    Parameters
    ----------
    telescope : `str`
        Which telescope convention to use for setting the conversion
        between rotTelPos and rotSkyPos. Default "rubin".
    """

    def __init__(self, telescope="rubin", nside=32):
        self.rc = rotation_converter(telescope=telescope)
        self.survey_features = {}

    def __call__(self, observation_list, conditions):
        # XXX--should I convert the list into an array and get rid of this
        # loop?
        for obs in observation_list:
            obs_pa, alt, az = pseudo_parallactic_angle(
                obs["RA"],
                obs["dec"],
                conditions.mjd,
                np.degrees(conditions.site.longitude_rad),
                np.degrees(conditions.site.latitude_rad),
            )

            obs["rotSkyPos"] = self.rc.rottelpos2rotskypos(0.0, obs_pa)

        return observation_list


class Comcam90rotDetailer(BaseDetailer):
    """
    Detailer to set the camera rotation so rotSkyPos is 0, 90, 180, or
    270 degrees. Whatever is closest to rotTelPos of zero.
    """

    def __init__(self, telescope="rubin", nside=32):
        self.rc = rotation_converter(telescope=telescope)
        self.survey_features = {}

    def __call__(self, observation_list, conditions):
        favored_rot_sky_pos = np.radians([0.0, 90.0, 180.0, 270.0, 360.0]).reshape(5, 1)
        obs_array = np.concatenate(observation_list)

        parallactic_angle, alt, az = np.radians(
            pseudo_parallactic_angle(
                obs_array["RA"],
                obs_array["dec"],
                conditions.mjd,
                np.degrees(conditions.site.longitude_rad),
                np.degrees(conditions.site.latitude_rad),
            )
        )

        # need to find the

        ang_diff = np.abs(self.rc._rotskypos2rottelpos(favored_rot_sky_pos, parallactic_angle))
        min_indxs = np.argmin(ang_diff, axis=0)
        # can swap 360 and zero if needed?
        final_rot_sky_pos = favored_rot_sky_pos[min_indxs]
        # Set all the observations to the proper rotSkyPos
        for rsp, obs in zip(final_rot_sky_pos, observation_list):
            obs["rotSkyPos"] = rsp

        return observation_list


class FixedSkyAngleDetailer(BaseDetailer):
    """Detailer to force a specific sky angle.

    Parameters
    ----------
    sky_angle : `float`, optional
        Desired sky angle (default = 0, in degrees).
    """

    def __init__(self, sky_angle=0.0, nside=32):
        super().__init__(nside=nside)

        self.sky_angle = np.radians(sky_angle)

    def __call__(self, observation_list, conditions):
        for observation in observation_list:
            observation["rotSkyPos"] = self.sky_angle

        return observation_list


class CloseAltDetailer(BaseDetailer):
    """
    re-order a list of observations so that the closest in altitude to
    the current pointing is first.

    Parameters
    ----------
    alt_band : `float` (10)
        The altitude band to try and stay in (degrees)
    """

    def __init__(self, alt_band=10.0):
        super(CloseAltDetailer, self).__init__()
        self.alt_band = IntRounded(np.radians(alt_band))

    def __call__(self, observation_list, conditions):
        obs_array = np.concatenate(observation_list)
        alt, az = _approx_ra_dec2_alt_az(
            obs_array["RA"],
            obs_array["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        alt_diff = np.abs(alt - conditions.tel_alt)
        in_band = np.where(IntRounded(alt_diff) <= self.alt_band)[0]
        if in_band.size == 0:
            in_band = np.arange(alt.size)

        # Find the closest in angular distance of the points that are in band
        ang_dist = _angular_separation(az[in_band], alt[in_band], conditions.tel_az, conditions.tel_alt)
        if np.size(ang_dist) == 1:
            good = 0
        else:
            good = np.min(np.where(ang_dist == ang_dist.min())[0])
        indx = in_band[good]
        result = observation_list[indx:] + observation_list[:indx]
        return result


class FlushForSchedDetailer(BaseDetailer):
    """Update the flush-by MJD to be before any scheduled observations

    Parameters
    ----------
    tol : `float`
         How much before to flush (minutes)
    """

    def __init__(self, tol=2.5):
        super(FlushForSchedDetailer, self).__init__()
        self.tol = tol / 24.0 / 60.0  # To days

    def __call__(self, observation_list, conditions):
        if np.size(conditions.scheduled_observations) > 0:
            new_flush = np.min(conditions.scheduled_observations) - self.tol
            for obs in observation_list:
                if obs["flush_by_mjd"] > new_flush:
                    obs["flush_by_mjd"] = new_flush
        return observation_list


class FilterNexp(BaseDetailer):
    """Demand one filter always be taken as a certain number of exposures"""

    def __init__(self, filtername="u", nexp=1, exptime=None):
        super(FilterNexp, self).__init__()
        self.filtername = filtername
        self.nexp = nexp
        self.exptime = exptime

    def __call__(self, observation_list, conditions):
        for obs in observation_list:
            if obs["filter"] == self.filtername:
                obs["nexp"] = self.nexp
                if self.exptime is not None:
                    obs["exptime"] = self.exptime
        return observation_list


class TakeAsPairsDetailer(BaseDetailer):
    def __init__(self, filtername="r", exptime=None, nexp_dict=None):
        """"""
        super(TakeAsPairsDetailer, self).__init__()
        self.filtername = filtername
        self.exptime = exptime
        self.nexp_dict = nexp_dict

    def __call__(self, observation_list, conditions):
        paired = copy.deepcopy(observation_list)
        if self.exptime is not None:
            for obs in paired:
                obs["exptime"] = self.exptime
        for obs in paired:
            obs["filter"] = self.filtername
            if self.nexp_dict is not None:
                obs["nexp"] = self.nexp_dict[self.filtername]
        if conditions.current_filter == self.filtername:
            for obs in paired:
                obs["scheduler_note"] = obs["scheduler_note"][0] + ", a"
            for obs in observation_list:
                obs["scheduler_note"] = obs["scheduler_note"][0] + ", b"
            result = paired + observation_list
        else:
            for obs in paired:
                obs["scheduler_note"] = obs["scheduler_note"][0] + ", b"
            for obs in observation_list:
                obs["scheduler_note"] = obs["scheduler_note"][0] + ", a"
            result = observation_list + paired
        return result


class TwilightTripleDetailer(BaseDetailer):
    def __init__(self, slew_estimate=5.0, n_repeat=3, update_note=True):
        super(TwilightTripleDetailer, self).__init__()
        self.slew_estimate = slew_estimate
        self.n_repeat = n_repeat
        self.update_note = update_note

    def __call__(self, observation_list, conditions):
        obs_array = np.concatenate(observation_list)

        # Estimate how much time is left in the twilgiht block
        potential_times = np.array(
            [
                conditions.sun_n18_setting - conditions.mjd,
                conditions.sun_n12_rising - conditions.mjd,
            ]
        )

        potential_times = np.min(potential_times[np.where(potential_times > 0)]) * 24.0 * 3600.0

        # How long will observations take?
        cumulative_slew = np.arange(obs_array.size) * self.slew_estimate
        cumulative_expt = np.cumsum(obs_array["exptime"])
        cumulative_time = cumulative_slew + cumulative_expt
        # If we are way over, truncate the list before doing the triple
        if np.max(cumulative_time) > potential_times:
            max_indx = np.where(cumulative_time / self.n_repeat <= potential_times)[0]
            if np.size(max_indx) == 0:
                # Very bad magic number fudge
                max_indx = 3
            else:
                max_indx = np.max(max_indx)
                if max_indx == 0:
                    max_indx += 1
            observation_list = observation_list[0:max_indx]

        # Repeat the observations n times
        out_obs = []
        for i in range(self.n_repeat):
            sub_list = copy.deepcopy(observation_list)
            if self.update_note:
                for obs in sub_list:
                    obs["scheduler_note"][0] += ", %i" % i
            out_obs.extend(sub_list)
        return out_obs



import unittest


class CleanUpTestCase(unittest.TestCase):
    def test_clean_up(self):
        """
        Test that sims_clean_up behaves as it should by importing a test module
        with some dummy caches, adding things to them, and then deleting them.
        """
        from test_modules.dummy_module import a_dict_cache, a_list_cache

        from rubin_scheduler.utils.code_utilities import sims_clean_up

        self.assertEqual(len(sims_clean_up.targets), 2)

        a_dict_cache["a"] = 1
        a_dict_cache["b"] = 2
        a_list_cache.append("alpha")
        a_list_cache.append("beta")

        self.assertEqual(len(a_dict_cache), 2)
        self.assertEqual(len(a_list_cache), 2)

        sims_clean_up()

        self.assertEqual(len(a_dict_cache), 0)
        self.assertEqual(len(a_list_cache), 0)
        self.assertEqual(len(sims_clean_up.targets), 2)

        # make sure that re-importing caches does not add second copies
        # to sims_clean_up.targets
        from test_modules.dummy_module import a_list_cache

        self.assertEqual(len(sims_clean_up.targets), 2)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.utils import SchemaConverter
from rubin_scheduler.utils.consdb import ConsDBVisits, load_consdb_visits


class TestConsdb(unittest.TestCase):

    @unittest.skip("avoid requiring access to consdb for tests.")
    def test_consdb_read_visits_quick(self):
        day_obs: str = "2024-06-26"
        consdb_visits: ConsDBVisits = load_consdb_visits("lsstcomcamsim", day_obs)
        schema_converter: SchemaConverter = SchemaConverter()

        opsim: pd.DataFrame = consdb_visits.opsim
        schema_converter.opsimdf2obs(opsim)

        obs: np.recarray = consdb_visits.obs
        schema_converter.obs2opsim(obs)
